%{
#define YYDEBUG 1
/* sdl_y.y
	yacc syntactical description of stack definition language, sdl.
*/

/* This pushdown automaton's stacktype: */
#define MAX_TEXT_LENGTH (512)
typedef struct {
	unsigned char value[512];
	unsigned char text[MAX_TEXT_LENGTH+1];
	unsigned char other[20];
	struct LinkNode *node;
	int stack_int;
} stack_block;
#define YYSTYPE stack_block

/* This is the size of a block of pointers
	to cards, used in a reallocation strategy
	for dynamic allocation. This number is
	used in both the card arrays for each stack
	in rpl_list, and for creating card arrays
	in WHEN-LINK constructions. */
#define CARD_PTR_BLOCK 500

/* This is for dynamic allocation of
	blocks of pointers to entries used in
	the WHEN-LINK constructions. */
#define ENTRY_PTR_BLOCK 500

/* Card */
struct Card {
	unsigned char name[50];
	int used;
	unsigned char *fields[50];
	struct Entry **entries;
} *current_card, *pushed_current_card;

/* Entry */
struct Entry {
	struct Card *card;
	int used;
	unsigned char *fields[50];
} *current_entry, *pushed_current_entry;

/* All the .rpl files, where
	the index is independent of the .rpl stack number,
	only indicating the order of creation. */
struct rpl {
	unsigned char fields[50][100];
	struct Card **cards;
	int card_index;
} *rpl_list[200];

/* The PAGE_TOTAL list keeps track of
	all assignments with PAGE_TOTAL on the value
	side, so that when the final number
	of cards in a card_book is finally determined,
	(in cases where the number of entries per card
	in data sensitive), it can be added to these
	variables.
*/
#define PAGE_TOTAL_LIST_LIMIT 10000
struct {
	struct Card *card;
	struct Entry *entry;
	int field;
} page_total_list[PAGE_TOTAL_LIST_LIMIT];
int ptl_index = 0;

/* Beginning of structures needed for 
	the WHEN-LINK constructions. */
struct Condition {
	int type;
	int field;
	int comparison;
	unsigned char value[100];
};

struct C_Condition {
	int typeA;
	int typeB;
	int fieldA;
	int fieldB;
	int comparison;
	unsigned char value[100];
};

struct {
	int entries;
	int cards;
	struct Entry **entry_list;
	struct Card **card_list;
	int list_index;
	struct Condition card_condition[20];
	struct Condition entry_condition[20];
	int rplnum_exists;
	int rplnum;
	int entry_condition_total;
	int card_condition_total;
	int c_condition_total;
	struct Condition c_condition[20];
} A,B;

struct LinkNode {
	int type;
	int list;
	int operator;
	struct LinkNode *left;
	struct LinkNode *right;
	int field;
	unsigned char value[100];
};

struct Link {
	int list;
	int field;
	int type;
	struct LinkNode *node;
};

struct {
	int condition_total;
	struct C_Condition conditions[20];
	int link_total;
	struct Link links[20];
} C;
/* End of structures needed for 
	the WHEN-LINK constructions. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

unsigned char *skip_list[20][100];
int skip_list_index[20];
struct Card *skip_card[100];
struct Entry *skip_entry[100];
struct Entry **skip_entries[100];
int skip_card_index[20];
int skip_card_num[20];
int starting_point;
int card_book_data_index;
unsigned char given_field_name_list[10][5][100];
unsigned char given_field_value_list[10][5][100];
unsigned char cb_field_name_list[10][100], 
	cb_field_value_list[10][100];

int repeat_data = 0;
int next_card = 0;
int lose_current_entry = 0;
int card_book_count = 1;
int pagetotal_seen = 0;
int pagetotal_enabled = 0;
int no_max_safeguard = 0;
int use_cif = 1;  /* Default behavior is to use the .cif file */

unsigned char *Ellip[] = {
(unsigned char *) "WGS84" ,
(unsigned char *) "BESSEL1841",
(unsigned char *) "HK1963",
(unsigned char *) "ED50",
(unsigned char *) "CLARK1880",
(unsigned char *) "WGS72",
(unsigned char *) "NAD27",
(unsigned char *) "GERMAN_FIX"};

#if 0
#define SD_DEBUG_LEX
#define SD_DEBUG1
#define SD_DEBUG2
#define SD_DEBUG3
#define SD_DEBUG4
#define SD_DEBUG5
#define SD_DEBUG6
#define SD_DEBUG7
#define FU_DEBUG
#define RICH_DEBUG
#endif

/* sorting #define SD_DEBUG4 */

#define EQUAL 1
#define NOT_EQUAL 2

#define ENTRY 1
#define CARD 2
#define STACK 3
#define OPERATOR 4
#define VALUE 5
#define MINUS 6
#define PLUS 7

int first_given[5] = {1,1,1,1,1};
int given_data_index[5];
int given_file_pos[5];
int given_line_pos[5];
int current_given = 0;
int skip_given = 0;

int u_card_num = 1;
int u_entry_total = 0;
int pushed_u_entry_total = 0;
int u_current_page = 0;
int u_page_total;

int first_book_card = 1;
int book_file_pos;
int book_line_pos;
int in_card_book = 0;

int in_book_entries = 0;
int entries_file_pos;
int entries_line_pos;
int max_book_entries;
int entries_ok = 0;

int file_lineno = 1;

int entry_index = 0;
int pushed_entry_index = 0;

extern unsigned char global_defs[25][50];
extern int globals_def_index;

extern struct {
	unsigned char name[50];
	unsigned char field_names[50][50];
	unsigned char list_field_names[50][50];
} card_types[100];
extern int card_type_index;

extern unsigned char field_names[100][100];
extern int field_name_index;

FILE *data_fd;
extern FILE *sdl_fd;

int assign_stack = 0;
struct Card * assign_card = (struct Card *) 0;
struct Entry * assign_entry = (struct Entry * ) 0;
int assign_name = 0;

unsigned char **data_ptr;
int current_repeat = 0;
int continue_repeat = 0;
int break_repeat = 0;
unsigned char sdl_file[50];
unsigned char data_file[50];
int data_size = 0;
int current_rpl = 0;
int rpl_index = 0;
unsigned char current_card_type[100];

int do_not_execute = 0;

unsigned char string_list[100][100];
int str_l_index = 0;
unsigned char sym_list[100][100];
int sym_l_index = 0;
int fd_input_done  = 0;
#define MAX_REPEATS 100
int data_lines = 0;

int in_new_stack = 0;
int in_new_card = 0;
int in_new_entry = 0;

#define ALPHA 1
#define REV_ALPHA 2
#define FORK 3
#define PAV 4
#define DATE 5
#define FLEUR 6
#define NUMERICAL 7
#define ALPHANUMERIC 8

#define MAX_CODE 100000
#define MAX_AFTER_CODE 5000000

#define SAFEGUARD_LIMIT 10000
/* above to prevent faults when programmer
	misuses max-entry-free cardbooks */

int data_records = 0;
int number_of_keys = 0;
int key_type[10];
unsigned char key_field[10][50];

#define MEM_ALLOCS 50000

#if 0
struct {
	unsigned char *ptr;
	int used;
} mems[MEM_ALLOCS+2];
#endif

int mem_tot = 0;
unsigned char *mem_temp;

unsigned char sdl_getc();
long repeat_line_pos[100];
long repeat_file_pos[100];
long p1=0;
long p2=0;
long p3=0;
long p4=0;

unsigned char *my_malloc(int size);
unsigned char *get_local_value(unsigned char *s);
struct Card *get_card(int rpl_number,int card_index);
struct Entry *get_entry(struct Card *card,int entry_index);
int data_field_index(unsigned char *field_name);
unsigned char *data_field(unsigned int index,unsigned int field);
unsigned char *itoa(int n,unsigned char *s);
void a_b_compare();
void clean_a_b();
void create_a();
void create_b();
void sort_a();
void sort_b();
void init_given_fields(stack_block *block);
void init_cb_fields(stack_block *block);
void sdl_restart ( FILE *input_file );
%}

%start sdl
%token NAME BEGIN_STACK REPEAT NEW_CARD NEW_ENTRY AFTER_STACK AFTER_ALL IF ELSE 
%token SORT_BY USE USE_RPL STRCPY PAGE END STRING NUMBER EXIST CONTINUE BREAK
%token SER_WRITE LNK_WRITE WPG_WRITE INDEX SUBSTR WHEN LINK STRCAT 
%token GIVEN_NEW CARD_BOOK BOOK_ENTRIES SKIP CIF_ON CIF_OFF
%token GT GTE LT LTE EQ NEQ STRCMP STRLEN ERROR QUAD NORMALIZE PRINT RICH_LINES
%token OR AND REPEAT_DATA NEXT_CARD LOSE_CURRENT_ENTRY TOC_WRITE FIND_UNIT SYM_SIZE
%left RPL_ASSIGN
%left OR
%left AND
%left '>' GTE '<' LTE EQ NEQ
%left '-' '+'
%left '*' '/'
%left UNARYMINUS '!' PLUS_PLUS MINUS_MINUS
%right TO_THE

%%

sdl: statements
		{
			printf("\nDone.\n");
		}
	;

statements: statement
	| statements statement
	;

statement: BEGIN_STACK 
		{
#			ifdef SD_DEBUG4
			yydebug = 1;
#			endif
			if (in_new_stack) {
				printf("You may not nest STACK blocks.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			in_new_stack = 1;
			if (!do_not_execute ) { 
				unsigned char buf[100];
				
				/* Making the first card in a stack. */
				current_card = (struct Card *) my_malloc(sizeof(struct Card));
				entry_index = 0;
				current_card->entries = (struct Entry **) my_malloc(100*sizeof(struct Entry *));
				current_card->entries[entry_index] = (struct Entry *) my_malloc(sizeof(struct Entry));
				current_entry = current_card->entries[entry_index];
				current_entry->card = current_card;
				entry_index++;

				/* Update the rpl_list to include this stack. */
				/* the rpl_list consists of pointers to all the
					rpl files so far constructed. These are
					all still in memory as linked lists. */
				rpl_list[rpl_index] = (struct rpl *) my_malloc(sizeof(struct rpl));
				rpl_list[rpl_index]->cards = (struct Card **) my_malloc(CARD_PTR_BLOCK*sizeof(struct Card *));
				rpl_list[rpl_index]->card_index = 1;
				rpl_list[rpl_index]->cards[rpl_list[rpl_index]->card_index] = current_card;
				current_rpl = rpl_index;
				rpl_index++;

				/* Card index maintenance. */
				/* note: rpl_list[current_rpl]->card_index is always one greater than the real current
					card number */
				rpl_list[current_rpl]->card_index++;
				u_card_num = 1;
				set_local_value("CARDNUM",itoa(u_card_num,buf));
			}
		}
		statements END
		{
			in_new_stack = 0;
		}
	| REPEAT
		{
			/* When the token REPEAT appears,
				the lexical module takes note
				of the file position and stores it in p1,
				whence it is pushed onto a stack
				of loop information here.
				This strategy is used for each loop
				interpretation.
			*/

			repeat_file_pos[current_repeat] = p1;
			repeat_line_pos[current_repeat++] = file_lineno;
		}
	'{' statements '}'
		{
			/* The state of do_not_execute is important for each
				interpretive instruction. Most only check it to see whether
				or not they should execute. Repeats always put themselves
				on the stack, whether or not they are executing, which doesn't
				seem correct but works...
				After a block is run, the value of do_not_execute might
				be CONTINUE, which implies that the loop be restarted, or
				BREAK which implies that it shouldn't. If the loop needs
				restarting, the stored file position is popped and the
				file repositioned.
			*/

			if (do_not_execute == CONTINUE && continue_repeat == current_repeat) {
				do_not_execute = 0;
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING ON from CONTINUE %d\n",current_repeat);
#				endif
			}
			if (!do_not_execute) { 
			  fseek(sdl_fd,repeat_file_pos[current_repeat-1],0);
			  file_lineno = repeat_line_pos[current_repeat-1];
			  // ### ber
			  sdl_flush_lex_buffers();
			  // sdl_restart(sdl_fd);
			  // yysptr = yysbuf;  /* Reset lexer input stream */
#				ifdef SD_DEBUG2
				printf("\nACTIVE REPEAT %d ENDS & STARTS ANEW\n",current_repeat);
#				endif
			}
			/* breaks and continues always effect only the correct level */ 
			if ((do_not_execute == BREAK) && (break_repeat == current_repeat) ) {
				do_not_execute = 0;
#				ifdef SD_DEBUG2
				printf("\nACTIVE REPEAT %d ENDS & BREAKS",current_repeat);
#				endif
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING ON from BREAK %d\n",current_repeat);
#				endif
			}
			current_repeat--;
		}
	| GIVEN_NEW '(' given_list ')'
		{
			/* GIVEN_NEW is a Piper-loop, so named because Jan Piper
				was going loopy trying to do what these structures do.
				They are loops with inherited conditions, so that the
				constraint on the outside loop (based on a changed
				value while traversing input records) applies also to
				the inner loop.
			*/

			/* if in card book, complain */
			if (in_card_book) {
				printf("You currently may not use GIVEN within a card_book.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			/* if in repeat, complain */
			if (current_repeat) {
				printf("You may not use GIVEN within repeat blocks.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (!do_not_execute) {
				if (first_given[current_given]) { 
					if (current_given) {
						given_data_index[current_given] = given_data_index[current_given-1];
					} else {
						given_data_index[current_given] = 0;
					}
					init_given_fields(&$3);
					first_given[current_given] = 0;
	
				}
				skip_card[current_given] = current_card;
				skip_entries[current_given] = current_card->entries;
				skip_entry[current_given] = current_entry;
				skip_card_index[current_given] = rpl_list[current_rpl]->card_index;
				skip_card_num[current_given] = u_card_num;

				given_file_pos[current_given] = p2;
				given_line_pos[current_given++] = file_lineno;
			}
		}
	'{' statements '}'
		{
			if ((do_not_execute == SKIP) && (skip_given == current_given)) {
				do_not_execute = 0;
				first_book_card = 1;
				given_undo();
			}

			/* if there's more new data start loop again */
			if (!do_not_execute) {
				if (given_advance(&$3)) { 
					fseek(sdl_fd,given_file_pos[current_given-1],0);
					file_lineno = given_line_pos[current_given-1];
					sdl_flush_lex_buffers();
					//sdl_restart(sdl_fd);
//			  		yysptr = yysbuf;  /* Reset lexer input stream */
				} else {
					first_given[current_given-1] = 1;
				}
	
				/* or fall out of loop */
				clear_skip_list();
				current_given--;
			}
		}
	| CARD_BOOK '(' book_list ')'
		{
			unsigned char buf[100];

			/* if in new card, complain */
			if (in_new_card) {
				printf("You currently may not call card_book within new_card.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			/* if in card book, complain */
			if (in_card_book) {
				printf("You currently may not nest card_books.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			in_card_book = 1;
			if (!do_not_execute) {

				/* user accessible variables */
				u_entry_total = 0;
				set_local_value("ENTRYTOTAL",itoa(u_entry_total,buf));
				u_current_page++;
				set_local_value("CURRENTPAGE",itoa(u_current_page,buf));
	
				if (first_book_card) {
					card_book_count = 1;
					set_u_page_total(&$3);
					if (current_given) {
						card_book_data_index = given_data_index[current_given -1];
					} else {
						card_book_data_index = 0;
					}
					/* Set Global User variable DATA_INDEX */
					set_local_value("DATA_INDEX",itoa(card_book_data_index,buf));
					init_cb_fields(&$3);
					first_book_card = 0;
					u_current_page = 1;
					set_local_value("CURRENTPAGE",itoa(u_current_page,buf));
				}
	
				/* for looping */
				book_file_pos = p3;
				book_line_pos = file_lineno;
			}
		}
	'{' statements '}'
		{
			int cba_return_value = 0;
			int new_value = 0;
			int old_value = 0;

			if (!do_not_execute) {
				make_new_card();
				/* if repeat_data, reset it, restart loop
					but don't do card_book_advance. */
				if (repeat_data)
				{
					repeat_data = 0;
					cba_return_value = 1;
				} else {
					cba_return_value = card_book_advance();
				}
				/* if there's more new data start loop again */
				if (entries_ok && cba_return_value) { 
					fseek( sdl_fd, book_file_pos, 0);
					file_lineno = book_line_pos;
					sdl_flush_lex_buffers();
					// sdl_restart(sdl_fd);
//			  		yysptr = yysbuf;  /* Reset lexer input stream */
					card_book_count++;
					printf("card_book_count incremented to %d\n",card_book_count);
				} else {
					/* fall out of loop */
					first_book_card = 1;

					/* put in pagetotals if any */
					/* ... if no skip */
					if (ptl_index > 0 && (max_book_entries == -1))
					{
						int i;
						unsigned char buf[100000];

						strcpy(buf,"");

						for (i=0;i<ptl_index;i++)
						{
							if (page_total_list[i].entry)
							{
								/* entry */
								if (page_total_list[i].entry->fields[page_total_list[i].field])
								{
									strcpy(buf,page_total_list[i].entry->fields[page_total_list[i].field]);
									new_value = atoi(buf) + card_book_count + 1;
									sprintf(buf,"%d",new_value);
									if (strlen(buf) > strlen(page_total_list[i].entry->fields[page_total_list[i].field]) + 2)
									{
										/* error & exit */
										printf("Failed to add pagetotal %d to entry fields after card book.\n",card_book_count);
										printf("Line %d. Exiting.\n",file_lineno);
										exit(1);
									}
									strcpy(page_total_list[i].entry->fields[page_total_list[i].field],buf);
									/* ... watch out for lose_entry entries! */
								}
							} else {
								/* card */
								if (page_total_list[i].card->fields[page_total_list[i].field])
								{
									strcpy(buf,page_total_list[i].card->fields[page_total_list[i].field]);
									old_value = atoi(buf);
									new_value = old_value + card_book_count + 1;
									printf("\nold_value %d, card_book_count %d, new_value %d\n",
										old_value, card_book_count, new_value); 
									sprintf(buf,"%d",new_value);
									if (strlen(buf) > (strlen(page_total_list[i].card->fields[page_total_list[i].field]) + 2))
									{
										/* error & exit */
										printf("Failed to add pagetotal %d to card fields after card book.\n",card_book_count);
										printf("Line %d. Exiting.\n",file_lineno);
										exit(1);
									}
									strcpy(page_total_list[i].card->fields[page_total_list[i].field],buf);
								}
							}

						}
						/* reset */
						ptl_index = 0;
						memset(page_total_list,0,sizeof(page_total_list));
						printf("card_book_count reset to %d\n",card_book_count);
					}
					card_book_count = 1;
				}
			}
			in_card_book = 0;
			if ((do_not_execute == BREAK)) {
				do_not_execute = 0;
				first_book_card = 1;
			}
		}
	| BOOK_ENTRIES
		{
			/* if not in card book, complain */
			if (!in_card_book) {
				printf("You must use book_entries within a card_book.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			/* if in book entries, complain */
			if (in_book_entries) {
				printf("You may not nest BOOK_ENTRIES blocks.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			in_book_entries = 1;
			if (!do_not_execute) {
				entries_ok = 1;

				entries_file_pos = p4;
				entries_line_pos = file_lineno;
			}
		}
	'{' statements '}'
		{
			unsigned char buf[100];
			int cba_return_value = 0;


			if (!do_not_execute) {
				/* if LOSE_CURRENT_ENTRY then do so. */
				if (lose_current_entry) 
				{
					lose_current_entry = 0;
					lose_entry();
					u_entry_total--;
				}
				make_new_entry();
				++u_entry_total;

				/* if under maximum entries, start loop again */
				/* or, if no max entries (-1), if NEXT_CARD not set and 
					no_max_safeguard is less than 10000. */
				if (max_book_entries == -1 && ++no_max_safeguard > SAFEGUARD_LIMIT) 
				{
					printf("%s: Bookentries loop iterated past safety limit %d. Line %d. Exiting.\n",
						sdl_file, SAFEGUARD_LIMIT, file_lineno);
					exit(1);
				}
				
				if ((max_book_entries == -1	&& next_card == 0) 
						|| (u_entry_total < max_book_entries)) 
				{
					if (repeat_data)
					{
						repeat_data = 0;
						cba_return_value = 1;
					} else {
						cba_return_value = card_book_advance();
					}
					if (cba_return_value)
					{
						fseek(sdl_fd,entries_file_pos,0);
						sdl_flush_lex_buffers();
						// sdl_restart(sdl_fd);
//			  			yysptr = yysbuf;  /* Reset lexer input stream */
						file_lineno = entries_line_pos;
					} else {
						entries_ok = 0; 
					}
				} else {
					/* otherwise fall out of loop */
					no_max_safeguard = 0;
	
					/* note: next_card's job is done just by not resetting book_entries
						loop. New card is automatically created when we fall out. */
					if (next_card) next_card = 0;
				}

				set_local_value("ENTRYTOTAL",itoa(u_entry_total,buf));
			}
			in_book_entries = 0;
		}
	| NEW_CARD '{' 
		{
			unsigned char buf[100];
#			ifdef SD_DEBUG5
			printf("\n NEW CARD number is %d.\n",rpl_list[current_rpl]->card_index-1);
#			endif
			if (in_new_card) {
				printf("You may not nest NEW_CARD blocks.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			in_new_card = 1;
			if (!do_not_execute) { 
				if (in_card_book) {
					make_and_push_card();
				}
				u_entry_total = 0;
				set_local_value("ENTRYTOTAL",itoa(u_entry_total,buf));
			}
		}
	statements '}'
		{
			if (!do_not_execute) { 
				if (in_card_book) {
					finish_and_pop_card();
				} else {
					make_new_card();
				}
			}
			in_new_card = 0;
		}
	| NEW_ENTRY '{' 
		{
			unsigned char buf[100];
			if (in_new_entry) {
				printf("You may not nest NEW_ENTRY blocks.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			in_new_entry = 1;
			if (!do_not_execute) { 
				u_entry_total++;
				set_local_value("ENTRYTOTAL",itoa(u_entry_total,buf));
			}
		}
		statements '}'
		{
			if (!do_not_execute) { 
				make_new_entry();
			}
			in_new_entry = 0;
		}
	| IF expr '{' 
		{
			/* everyone watches do_not_execute */
			if ((!do_not_execute) && (!atoi($2.value))) {
				do_not_execute = 1;
				strcpy($$.text,"DO_NOT_EXECUTE");
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING OFF from IF\n");
#				endif
			}
		}
		statements '}'
		{
			/* if execution turned off by partner, turn on. */
			if (!strcmp($4.text,"DO_NOT_EXECUTE")) {
				do_not_execute = 0;	
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING ON from IF\n");
#				endif
			}
		}
	| WHEN '{' when_conditions '}' LINK '{' link_statements '}'
		{
			/* don't do within a card, entry or a repeat construct */
			if (in_card_book) {
				printf("You may not use WHEN-LINK within a card_book.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (in_new_card) {
				printf("You may not use WHEN-LINK in a card block.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (in_new_entry) {
				printf("You may not use WHEN-LINK in an entry block.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (in_book_entries) {
				printf("You may not use WHEN-LINK in a book_entries block.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (current_repeat) {
				printf("You may not use WHEN-LINK in a repeat block.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (current_given) {
				printf("You may not use WHEN-LINK in a given_new block.\n");
				printf("Line %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (!do_not_execute) {
				/* (first we check for (a,b) type sets already 
					restricted by the given (a,b) conditions,
					and sorted by the given (c) keys in the
					proper order.) */
				create_a();
				create_b();
				sort_a();
				sort_b();
				a_b_compare();
				clean_a_b();
			}

		}
	| NAME '=' expr ';'
		{
			/* update or create local variable. */
			if (!do_not_execute) {
				if ($1.text[0] == '_') {
					printf("\nSDL: Used _variable in '=' expression.\n");
					printf("Perhaps use -> instead.\n");
					printf("Variable: %s. Line %d. Exiting.\n",
						$1.text,file_lineno);
					exit(1);
				}
				set_local_value($1.text,$3.value);
#				ifdef SD_DEBUG3
				printf("#### %s is %s.",$1.text,$3.value);
#				endif
			}
		}
	| SORT_BY '(' sort_by_args ')' ';'
		{
			int i;

			if (!do_not_execute) { 
				if (!strlen(data_file)) {
					printf("%s: sort before data available. Line %d. Exiting.\n",
						sdl_file, file_lineno);
					exit(1);
				}
				sort();
#				ifdef SD_DEBUG6
				for (i=0;i<data_records;i++)
					printf("%s\n",data_ptr[i]); 
#				endif
				erase_keys();
			}
		}
	| USE '(' expr ')' ';'
		{
			int i;
			if (!do_not_execute) { 
				if (strlen(data_file)) {
					for (i=0;i<data_lines;i++){
						my_free(data_ptr[i]);
					}
					my_free(data_ptr);
					fclose(data_fd);
				}
				strcpy(data_file,$3.value);
				if ((data_fd=fopen(data_file,"r"))==NULL) {
					printf("%s: couldn't open %s. Line %d. Exiting.\n",
						sdl_file,data_file,file_lineno);
					exit(1);
				}
				data_size = data_file_size(data_fd);	
				if ((data_ptr = (unsigned char **) my_malloc((data_lines+4)*(sizeof(unsigned char *))))
					== NULL) {
					printf("SDL: Out of Memory!\n");
					printf("after opening %s. Exiting. \n", $3.value);
					exit(1);
				}
				for (i=0;i<data_lines+4;i++){
					data_ptr[i] = my_malloc(1000);
				}
				read_data(data_ptr,data_fd);
			}
		}
	| PRINT '(' expr ')' ';'
		{
			if (!do_not_execute) { 
				printf("%s\n",$3.value);
				fflush(stdout);
			}
		}
	| ERROR '(' STRING ')' ';'
		{
			if (!do_not_execute) {
				printf("*** User called error\n");
				yyerror($3.value);
			}
		}
	| NAME MINUS_MINUS ';'
		{
			unsigned char *value;

			if (!do_not_execute) { 
				value = get_local_value($1.text);
				itoa((atoi(value)-1),value);
				set_local_value($1.text,value);
#				ifdef SD_DEBUG3
				printf("#### -- Now %s is %s.",$1.text,value);
#				endif
			}
		}
	| NAME PLUS_PLUS ';'
		{
			unsigned char *value;

			if (!do_not_execute) { 
				value = get_local_value($1.text);
				itoa((atoi(value)+1),value);
				set_local_value($1.text,value);
#				ifdef SD_DEBUG3
				printf("#### ++ Now %s is %s.",$1.text,value);
#				endif
			}
		}
	| PAGE '(' NAME ',' NAME ',' expr ')' ';'
		{
			unsigned char *buf1;
			unsigned char *buf2;

			/* extract current and total pages */
			if (!do_not_execute) { 
				buf1 = my_malloc(100);
				buf2 = my_malloc(100);
				page(buf1,buf2,$7.value);
				set_local_value($3.text,buf1);
				set_local_value($5.text,buf2);
				my_free(buf1);
				my_free(buf2);
			}
		}
	| STRCPY '(' NAME ',' expr ')' ';'
		{
			unsigned char *buf;

			if (!do_not_execute) { 
				buf = my_malloc(500);
				strcpy(buf,$5.value);
				set_local_value($3.text,buf);
				my_free(buf);
			}
		}
	| STRCAT '(' NAME ',' expr ')' ';'
		{
			unsigned char *buf;

			if (!do_not_execute) { 
				buf = my_malloc(500);
				strcpy(buf,get_local_value($3.text));
				strcat(buf,$5.value);
				set_local_value($3.text,buf);
				my_free(buf);
			}
		}
	| CONTINUE ';'
		{
			if (!do_not_execute) { 
				do_not_execute = CONTINUE;
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING OFF from CONTINUE %d\n",current_repeat);
#				endif
				continue_repeat = current_repeat;
			}
		}
	| BREAK ';'
		{
			if (!do_not_execute) {
				do_not_execute = BREAK;
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING OFF from BREAK %d\n",current_repeat);
#				endif
				break_repeat = current_repeat;
			}

		}
	| SKIP ';'
		{
			if (!do_not_execute) {
				/* if not in card book, complain */
				if (!current_given) {
					printf("You must use SKIP within GIVEN_NEW block.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				do_not_execute = SKIP;
#				ifdef SD_DEBUG3
				printf("\n***TURNING EVERYTHING OFF from BREAK %d\n",current_repeat);
#				endif
				skip_given = current_given;
			}

		}
	
	| REPEAT_DATA ';'
		{
			/* if not in book_entries, complain */
			if (!do_not_execute) {
				if (!in_book_entries) {
					printf("You must use repeat_data within book_entries.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				repeat_data = 1;
			}
		}

	| NEXT_CARD ';'
		{
			/* if not in book_entries, complain */
			if (!do_not_execute) {
				if (!in_book_entries) {
					printf("You must use next_card within book_entries.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				next_card = 1;
			}
		}

	| LOSE_CURRENT_ENTRY ';'
		{
			if (!do_not_execute) 
			{
				if (!in_book_entries) {
					printf("You must use lose_current_entry within book_entries.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				lose_current_entry = 1;
			}
		}
	| NAME '[' expr ']' '[' expr ']' '[' expr ']' RPL_ASSIGN expr ';'
		{
			unsigned char *ptr;
			int len;

			if (!do_not_execute) {
				assign_stack = get_rpl_number($3.value);
				assign_card = get_card(assign_stack,atoi($6.value));
				assign_entry = get_entry(assign_card,atoi($9.value));
				assign_name = entry_field_index($1.text,assign_card->name);
				len = strlen($12.value);
				ptr = my_malloc(len+2);
				assign_entry->fields[assign_name] = ptr;
				strcpy(assign_entry->fields[assign_name],$12.value);
				assign_entry->used = 1;
				assign_entry = 0;
			}
		}
	| NAME '[' expr ']' '[' expr ']' RPL_ASSIGN expr ';'
		{
			unsigned char *ptr;
			int len;

			/* ... exception for _Card needs to be made (not a field) */
			if (!do_not_execute) {
				assign_stack = get_rpl_number($3.value);
				assign_card = get_card(assign_stack,atoi($6.value));
				assign_name = card_field_index($1.text,assign_card->name);
				len = strlen($9.value);
				ptr = my_malloc(len+2);
				assign_card->fields[assign_name] = ptr;
				strcpy(assign_card->fields[assign_name],$9.value);
				assign_card->used = 1;
			}
		}
	| NAME RPL_ASSIGN 
		{
			if (!do_not_execute) { 
				pagetotal_enabled = 1;	
				pagetotal_seen = 0;
			}
		}
		expr ';'
		{
			unsigned char *ptr;
			int len,field_index;


			/* this is the only rpl_assign that anyone uses anymore. */
			if (!do_not_execute) { 
				pagetotal_enabled = 0;
				if ($1.text[0] == '_') {
					if (!in_new_card && !in_new_entry && !in_card_book) {
							printf("\nSDL: _Assignments without index must be made in\n");
							printf("	new_card, card_book or new_entry block.\n");
							printf("	Line %d. Exiting.\n",file_lineno);
							exit(1);
					}
					if ($1.text[1] == '_') {
						/* entry field */
						field_index = entry_field_index(&$1.text[2],
							current_card_type);
						len = strlen($4.value);
						ptr = my_malloc(len+4);
						current_entry->fields[field_index] = ptr;
						strcpy(current_entry->fields[field_index],$4.value);
						current_entry->used = 1;
					} else {
						/* card field */
						if (!strcmp($1.text,"_Card")) {
							strcpy(current_card_type,$4.value);
							strcpy(current_card->name,$4.value);
							current_card->used = 1;
						} else {
							field_index = card_field_index(&$1.text[1],
								current_card_type);
							len = strlen($4.value);
							ptr = my_malloc(len+4);
							current_card->fields[field_index] = ptr;
							strcpy(current_card->fields[field_index],$4.value);
							current_card->used = 1;
						}
					}
				} else {
					/* .rpl global */
					field_index = global_field_index($1.text);
					strcpy(rpl_list[current_rpl]->fields[field_index],$4.value);
				}
				/* Save this card, entry and field. */
				if (in_card_book && pagetotal_seen && (max_book_entries == -1))
				{
					pagetotal_seen = 0;

					if ($1.text[0] == '_') {
						if ($1.text[1] == '_') {
#							ifdef SD_DEBUG7
							printf("\nPAGETOTAL entry current_card %d, current_entry %d, field_index %d\n",
								current_card,current_entry,field_index);
#							endif
							if (!current_entry->fields[field_index])
							{
								printf("There is no pointer at entry field %s.\n",$1.text);
								printf("Line %d. Exiting.\n",file_lineno);
								exit(1);
							}
							/* entry */
							page_total_list[ptl_index].card = current_card;
							page_total_list[ptl_index].entry = current_entry;
							page_total_list[ptl_index++].field = field_index;
							if (ptl_index > PAGE_TOTAL_LIST_LIMIT)
							{
								printf("The number of page_totals to be inserted exceeded PAGE_TOTAL_LIST_LIMIT %d.\n",PAGE_TOTAL_LIST_LIMIT);
								printf("Line %d. Exiting.\n",file_lineno);
								exit(1);
							}
						} else {
#							ifdef SD_DEBUG7
							printf("\nPAGETOTAL card current_card %d, current_entry %d, field_index %d\n",
								current_card,current_entry,field_index);
#							endif
							if (!current_card->fields[field_index])
							{
								printf("There is no pointer at card field %s.\n",$1.text);
								printf("Line %d. Exiting.\n",file_lineno);
								exit(1);
							}
							/* card */
							page_total_list[ptl_index].card = current_card;
							/* entry is 0 if it's a card field */
							page_total_list[ptl_index].entry = 0;
							page_total_list[ptl_index++].field = field_index;
							if (ptl_index > PAGE_TOTAL_LIST_LIMIT)
							{
								printf("The number of page_totals to be inserted exceeded PAGE_TOTAL_LIST_LIMIT %d.\n",PAGE_TOTAL_LIST_LIMIT);
								printf("Line %d. Exiting.\n",file_lineno);
								exit(1);
							}
						}
					}
				}
			}
		}
	;

var: 	NAME '[' expr ']' '[' expr ']' '[' expr ']'
		{
			/* get value of entry field given rpl card, and entry, e.g. "S2Offset [21][15][2]" */
			if (!do_not_execute) { 
				assign_stack = get_rpl_number($3.value);

#				ifdef SD_DEBUG5
				printf("\n#### rpl_number for %s is %d.",$3.value,assign_stack);
#				endif

				assign_card = get_card(assign_stack,atoi($6.value));

#				ifdef SD_DEBUG5
				printf("\n#### card %s for rpl %d is %d.",$6.value,assign_stack,assign_card);
#				endif

				assign_entry = get_entry(assign_card,atoi($9.value));
				assign_name = entry_field_index($1.text,assign_card->name);
				if (!assign_entry->fields[assign_name]) {
					strcpy($$.value,"");
				} else {
					strcpy($$.value,assign_entry->fields[assign_name]);
				}

#				ifdef SD_DEBUG5
				printf("\n#### overall value is %s.\n",$$.value);
#				endif
				assign_entry = 0;
			}
		}
	| NAME '[' expr ']' '[' expr ']'
		{
			/* get value of card field given rpl and card, e.g. "City[12][100]" */
			if (!do_not_execute) { 
				assign_stack = get_rpl_number($3.value);

#				ifdef SD_DEBUG5
				printf("\n#### rpl_number for %s is %d.",$3.value,assign_stack);
#				endif

				assign_card = get_card(assign_stack,atoi($6.value));

#				ifdef SD_DEBUG5
				printf("\n#### card %s for rpl %d is %d.",$6.value,assign_stack,assign_card);
#				endif

				if (!strcasecmp($1.text,"card")){

#					ifdef SD_DEBUG5
					printf("\n#### card %s's name is %s.",assign_card,assign_card->name);
#					endif

					strcpy($$.value,assign_card->name);
				} else {
					assign_name = card_field_index($1.text,assign_card->name);
					if (!assign_card->fields[assign_name]) {
						strcpy($$.value,"");
					} else {
						strcpy($$.value,assign_card->fields[assign_name]);
					}
	
#					ifdef SD_DEBUG5
					printf("\n#### overall value is %s.\n",$$.value);
#					endif
				}
			}

		}
	| NAME '[' expr ']'
		{
			/* this referes to source data, not .rpl (output) data */
			int field_index;
			unsigned char *ptr;
			
			if (!do_not_execute) { 
				if ($1.text[0] == '$') {
					/*  index to the active read structure. */
					field_index = data_field_index(&$1.text[1]);
					strcpy($$.value,(ptr=data_field( atoi($3.value), field_index)));
					my_free(ptr);
#					ifdef SD_DEBUG3
					printf("#### %s [ %s ] is %s.",$1.text,$3.value,$$.value);
#					endif
				} else {
					printf("SDL: name with single index used without $ prefix.\n");
					printf("Variable %s. Line %d. Exiting.\n",$1.text,file_lineno);
					exit(1);
				}
			}
		}
	| NAME '[' ']'
		{
			/* this referes to source data, not .rpl (output) data */
			int field_index;
			unsigned char *ptr;
			
			if (!do_not_execute) {
				/* complain if not in given or card_book */
				if (!in_card_book && !current_given) {
					printf("You may not use empty brackets outside card_book or given_new.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				if ($1.text[0] == '$') {
					/*  index to the active read structure. */
					field_index = data_field_index(&$1.text[1]);
					if (in_card_book) {
						strcpy($$.value,(ptr=data_field(card_book_data_index, field_index)));
					} else {
						strcpy($$.value,(ptr=data_field(given_data_index[current_given-1], field_index)));
					}
					my_free(ptr);
				} else {
					printf("SDL: name with blank index used without $ prefix.\n");
					printf("Variable %s. Line %d. Exiting.\n",$1.text,file_lineno);
					exit(1);
				}
			}
		}
	| NAME
		{
			/* check for badly placed names */
			if (!do_not_execute) {
				if (pagetotal_enabled && in_card_book && !strcmp($1.text,"PAGETOTAL"))
				{
					pagetotal_seen = 1;
				}
				if ($1.text[0] == '$') {
					printf("SDL: source reference (using $) must have index.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				if ($1.text[0] == '_') {
					printf("SDL: _var should only be used on left side of '<-' .\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				/* user symbol */
				strcpy($$.value,get_local_value($1.text));
#				ifdef SD_DEBUG3
				printf("##### %s is %s.",$1.text,$$.value);
#				endif
			}
		}
	;

sort_by_args: NAME ',' NAME
		{
			/* write into keys */
			if (!do_not_execute) {
				key_type[number_of_keys] = string_to_key($1.text);
				strcpy(key_field[number_of_keys],$3.text);
				number_of_keys++;
			}
		}
	| sort_by_args ',' NAME ',' NAME
		{
			/* write into keys */
			if (!do_not_execute) {
				key_type[number_of_keys] = string_to_key($3.text);
				strcpy(key_field[number_of_keys],$5.text);
				number_of_keys++;
			}
		}
	;

expr: 	 expr OR expr	
		{ 
			if (!do_not_execute) {
				if (atoi($1.value) || atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr AND expr 
		{ 
			if (!do_not_execute) {
				if (atoi($1.value) && atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| '!' expr 	
		{ 
			if (!do_not_execute) {
				if (!atoi($2.value)) strcpy($$.value,"1"); else strcpy($$.value,"0"); 
			}
		}
	| expr '>' expr 	
		{ 
			if (!do_not_execute) {
				if (atoi($1.value) > atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr GTE expr 
		{ 
			if (!do_not_execute) {
				if (atoi($1.value)>= atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr '<' expr 	
		{ 
			if (!do_not_execute) {
				if (atoi($1.value) < atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr LTE expr 
		{ 
			if (!do_not_execute) {
				if (atoi($1.value)<= atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr EQ expr 	
		{ 
			if (!do_not_execute) {
				if (atoi($1.value) == atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr NEQ expr 
		{ 
			if (!do_not_execute) {
				if (atoi($1.value) != atoi($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0");
			}
		}
	| expr '-' expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) - atoi($3.value)),$$.value); 
			}
		}
	| expr '+' expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) + atoi($3.value)),$$.value); 
			}
		}
	| expr '*' expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) * atoi($3.value)),$$.value); 
			}
		}
	| expr '/' expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) / atoi($3.value)),$$.value); 
			}
		}
	| '(' expr ')' 	
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$2.value); 
			}
		}
	| expr TO_THE expr 		
		{ 
			if (!do_not_execute) {
				itoa(pow(atoi($1.value),atoi($3.value)),$$.value); 
			}
		}
	| '-' expr %prec UNARYMINUS 	
		{ 
			if (!do_not_execute) {
				itoa((atoi($2.value) * -1),$$.value); 
			}
		}
	| var 			
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$1.value); 
				strcpy($$.text,$1.text);
			}
		}
	| NUMBER 		
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$1.value); 
				strcpy($$.text,$1.text);
			}
		}
	| EXIST '(' expr ')' 	
		{ 
			if (!do_not_execute) {
				if ($3.value)
					{
					if (strlen($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0"); 
					}
				else
					{
					strcpy($$.value,"0");
					}
			}
		}
	| STRCMP '(' expr ',' expr ')' 
		{ 
			if (!do_not_execute) {
				itoa(strcmp($3.value,$5.value),$$.value); 
			}
		}
	| SUBSTR '(' expr ',' expr ',' expr ')' 
		{ 
			if (!do_not_execute) {
				strncpy($$.value,&$3.value[atoi($5.value)],atoi($7.value)); 
				$$.value[atoi($7.value)] = '\0';
			}
		}
	| INDEX '(' expr ',' expr ')' 
		{ 
			if (!do_not_execute) {
				itoa(myindex($3.value,$5.value),$$.value); 
			}
		}
	| STRLEN '(' expr ')' 
		{ 
			if (!do_not_execute) {
				itoa(strlen($3.value),$$.value); 
			}
		}
	| STRING		
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$1.value); 
				strcpy($$.text,$1.text);
			}
		}
	;

when_expr: when_expr '-' when_expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) - atoi($3.value)),$$.value); 
			}
		}
	| when_expr '+' when_expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) + atoi($3.value)),$$.value); 
			}
		}
	| when_expr '*' when_expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) * atoi($3.value)),$$.value); 
			}
		}
	| when_expr '/' when_expr 
		{ 
			if (!do_not_execute) {
				itoa((atoi($1.value) / atoi($3.value)),$$.value); 
			}
		}
	| '(' when_expr ')' 	
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$2.value); 
			}
		}
	| when_expr TO_THE when_expr 		
		{ 
			if (!do_not_execute) {
				itoa(pow(atoi($1.value),atoi($3.value)),$$.value); 
			}
		}
	| '-' when_expr %prec UNARYMINUS 	
		{ 
			if (!do_not_execute) {
				itoa((atoi($2.value) * -1),$$.value); 
			}
		}
	| var 			
		{
			if (!do_not_execute) {
				strcpy($$.value,$1.value); 
				strcpy($$.text,$1.text);
			}
		}
	| NUMBER 		
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$1.value); 
				strcpy($$.text,$1.text);
			}
		}
	| EXIST '(' when_expr ')' 	
		{ 
			if (!do_not_execute) {
				if (strlen($3.value)) strcpy($$.value,"1"); else strcpy($$.value,"0"); 
			}
		}
	| STRCMP '(' when_expr ',' when_expr ')' 
		{ 
			if (!do_not_execute) {
				itoa(strcmp($3.value,$5.value),$$.value); 
			}
		}
	| SUBSTR '(' when_expr ',' when_expr ',' when_expr ')' 
		{ 
			if (!do_not_execute) {
				strncpy($$.value,&$3.value[atoi($5.value)],atoi($7.value)); 
				$$.value[atoi($7.value)] = '\0';
			}
		}
	| INDEX '(' when_expr ',' when_expr ')' 
		{ 
			if (!do_not_execute) {
				itoa(myindex($3.value,$5.value),$$.value); 
			}
		}
	| STRLEN '(' when_expr ')' 
		{ 
			if (!do_not_execute) {
				itoa(strlen($3.value),$$.value); 
			}
		}
	| STRING		
		{ 
			if (!do_not_execute) {
				strcpy($$.value,$1.value); 
				strcpy($$.text,$1.text);
			}
		}
	;

when_conditions: when_condition
	| when_conditions ',' when_condition
	;

when_condition: when_value EQ when_value 	
		{
			/* determine condition type (a, b) or c based on what's in when_values 
			for type (a,b), put condition on list for that set,
			which will be used upon close of when section to 
			narrow down the data for the linking. For type c, put 
			encoded condition on list for link time */

			if (!do_not_execute) {
			/* condition for A */
			if ((!strcmp($1.text,"A")) && (!strcmp($3.text,"VALUE"))) {
				if (!strcmp($1.value,"ENTRY")) {
					A.entry_condition[A.entry_condition_total].field = A_entry_lookup($1.other);
					strcpy(A.entry_condition[A.entry_condition_total].value,$3.value);
					A.entry_condition[A.entry_condition_total].comparison = EQUAL;
					A.entry_condition[A.entry_condition_total].type = ENTRY;
					A.entry_condition_total++;
				} 
				if (!strcmp($1.value,"CARD")) {
					if (!strcmp($1.other,"Card")) {
						A.card_condition[A.card_condition_total].field = -1;
					} else {
						A.card_condition[A.card_condition_total].field = A_card_lookup($1.other);
					}
					strcpy(A.card_condition[A.card_condition_total].value,$3.value);
					A.card_condition[A.card_condition_total].comparison = EQUAL;
					A.card_condition[A.card_condition_total].type = CARD;
					A.card_condition_total++;
				}
			}
			if ((!strcmp($1.text,"VALUE")) && (!strcmp($3.text,"A"))) {
				if (!strcmp($3.value,"ENTRY")) {
					A.entry_condition[A.entry_condition_total].field = A_entry_lookup($3.other);
					strcpy(A.entry_condition[A.entry_condition_total].value,$1.value);
					A.entry_condition[A.entry_condition_total].comparison = EQUAL;
					A.entry_condition[A.entry_condition_total].type = ENTRY;
					A.entry_condition_total++;
				} 
				if (!strcmp($3.value,"CARD")) {
					if (!strcmp($3.other,"Card")) {
						A.card_condition[A.card_condition_total].field = -1;
					} else {
						A.card_condition[A.card_condition_total].field = A_card_lookup($3.other);
					}
					strcpy(A.card_condition[A.card_condition_total].value,$1.value);
					A.card_condition[A.card_condition_total].comparison = EQUAL;
					A.card_condition[A.card_condition_total].type = CARD;
					A.card_condition_total++;
				}
			}


			/* condition for B */
			if ((!strcmp($1.text,"B")) && (!strcmp($3.text,"VALUE"))) {
				if (!strcmp($1.value,"ENTRY")) {
					B.entry_condition[B.entry_condition_total].field = B_entry_lookup($1.other);
					strcpy(B.entry_condition[B.entry_condition_total].value,$3.value);
					B.entry_condition[B.entry_condition_total].comparison = EQUAL;
					B.entry_condition[B.entry_condition_total].type = ENTRY;
					B.entry_condition_total++;
				} 
				if (!strcmp($1.value,"CARD")) {
					if (!strcmp($1.other,"Card")) {
						B.card_condition[B.card_condition_total].field = -1;
					} else {
						B.card_condition[B.card_condition_total].field = B_card_lookup($1.other);
					}
					strcpy(B.card_condition[B.card_condition_total].value,$3.value);
					B.card_condition[B.card_condition_total].comparison = EQUAL;
					B.card_condition[B.card_condition_total].type = CARD;
					B.card_condition_total++;
				}
			}
			if ((!strcmp($1.text,"VALUE")) && (!strcmp($3.text,"B"))) {
				if (!strcmp($3.value,"ENTRY")) {
					B.entry_condition[B.entry_condition_total].field = B_entry_lookup($3.other);
					strcpy(B.entry_condition[B.entry_condition_total].value,$1.value);
					B.entry_condition[B.entry_condition_total].comparison = EQUAL;
					B.entry_condition[B.entry_condition_total].type = ENTRY;
					B.entry_condition_total++;
				} 
				if (!strcmp($3.value,"CARD")) {
					if (!strcmp($3.other,"Card")) {
						B.card_condition[B.card_condition_total].field = -1;
					} else {
						B.card_condition[B.card_condition_total].field = B_card_lookup($3.other);
					}
					strcpy(B.card_condition[B.card_condition_total].value,$1.value);
					B.card_condition[B.card_condition_total].comparison = EQUAL;
					B.card_condition[B.card_condition_total].type = CARD;
					B.card_condition_total++;
				}
			}

			if ((!strcmp($1.text,"B")) && (!strcmp($3.text,"A"))) {

				/* a,b condition -- for sort */
				if (!strcmp($1.value,"ENTRY")) {
				        B.c_condition[B.c_condition_total].field = B_entry_lookup($1.other);
					B.c_condition[B.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeA = ENTRY;
					C.conditions[C.condition_total].fieldA = B_entry_lookup($1.other);
				}
				if (!strcmp($1.value,"CARD")) {
				        B.c_condition[B.c_condition_total].field = B_card_lookup($1.other);
					B.c_condition[B.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeA = CARD;
					C.conditions[C.condition_total].fieldA = B_card_lookup($1.other);
				}
				if (!strcmp($3.value,"ENTRY")) {
				        A.c_condition[A.c_condition_total].field = A_entry_lookup($3.other);
					A.c_condition[A.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeB = ENTRY;
					C.conditions[C.condition_total].fieldB = A_entry_lookup($3.other);
				}
				if (!strcmp($3.value,"CARD")) {
				        A.c_condition[A.c_condition_total].field = A_card_lookup($3.other);
					A.c_condition[A.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeB = CARD;
					C.conditions[C.condition_total].fieldB = A_card_lookup($3.other);
				}
				A.c_condition_total++;
				B.c_condition_total++;

				/* a,b condition -- for cartesian product search */
				C.conditions[C.condition_total].comparison = EQUAL;
				C.condition_total++;

			}

			if ((!strcmp($1.text,"A")) && (!strcmp($3.text,"B"))) {
				if (!strcmp($3.value,"ENTRY")) {
				        B.c_condition[B.c_condition_total].field = B_entry_lookup($3.other);
					B.c_condition[B.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeB = ENTRY;
					C.conditions[C.condition_total].fieldB = B_entry_lookup($3.other);
				}
				if (!strcmp($3.value,"CARD")) {
				        B.c_condition[B.c_condition_total].field = B_card_lookup($3.other);
					B.c_condition[B.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeB = CARD;
					C.conditions[C.condition_total].fieldB = B_card_lookup($3.other);
				}
				if (!strcmp($1.value,"ENTRY")) {
				        A.c_condition[A.c_condition_total].field = A_entry_lookup($1.other);
					A.c_condition[A.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeA = ENTRY;
					C.conditions[C.condition_total].fieldA = A_entry_lookup($1.other);
				}
				if (!strcmp($1.value,"CARD")) {
				        A.c_condition[A.c_condition_total].field = A_card_lookup($1.other);
					A.c_condition[A.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeA = CARD;
					C.conditions[C.condition_total].fieldA = A_card_lookup($1.other);
				}
				A.c_condition_total++;
				B.c_condition_total++;

				/* a,b condition -- for cartesian product search */
				C.conditions[C.condition_total].comparison = EQUAL;
				C.condition_total++;
			}
			}
		}
	| when_value NEQ when_value 
		{
			/* determine condition type (a, b) or c based on what's in when_values 
			for type (a,b), put condition on list for that set,
			which will be used upon close of when section to 
			narrow down the data for the linking. For type c, put 
			encoded condition on list for link time */

			if (!do_not_execute) {
			/* condition for A */
			if ((!strcmp($1.text,"A")) && (!strcmp($3.text,"VALUE"))) {
				if (!strcmp($1.value,"ENTRY")) {
					A.entry_condition[A.entry_condition_total].field = A_entry_lookup($1.other);
					strcpy(A.entry_condition[A.entry_condition_total].value,$3.value);
					A.entry_condition[A.entry_condition_total].comparison = NOT_EQUAL;
					A.entry_condition[A.entry_condition_total].type = ENTRY;
					A.entry_condition_total++;
				} 
				if (!strcmp($1.value,"CARD")) {
					if (!strcmp($1.other,"Card")) {
						A.card_condition[A.card_condition_total].field = -1;
					} else {
						A.card_condition[A.card_condition_total].field = A_card_lookup($1.other);
					}
					strcpy(A.card_condition[A.card_condition_total].value,$3.value);
					A.card_condition[A.card_condition_total].comparison = NOT_EQUAL;
					A.card_condition[A.card_condition_total].type = CARD;
					A.card_condition_total++;
				}
			}
			if ((!strcmp($1.text,"VALUE")) && (!strcmp($3.text,"A"))) {
				if (!strcmp($3.value,"ENTRY")) {
					A.entry_condition[A.entry_condition_total].field = A_entry_lookup($3.other);
					strcpy(A.entry_condition[A.entry_condition_total].value,$1.value);
					A.entry_condition[A.entry_condition_total].comparison = NOT_EQUAL;
					A.entry_condition[A.entry_condition_total].type = ENTRY;
					A.entry_condition_total++;
				} 
				if (!strcmp($3.value,"CARD")) {
					if (!strcmp($3.other,"Card")) {
						A.card_condition[A.card_condition_total].field = -1;
					} else {
						A.card_condition[A.card_condition_total].field = A_card_lookup($3.other);
					}
					strcpy(A.card_condition[A.card_condition_total].value,$1.value);
					A.card_condition[A.card_condition_total].comparison = NOT_EQUAL;
					A.card_condition[A.card_condition_total].type = CARD;
					A.card_condition_total++;
				}
			}


			/* condition for B */
			if ((!strcmp($1.text,"B")) && (!strcmp($3.text,"VALUE"))) {
				if (!strcmp($1.value,"ENTRY")) {
					B.entry_condition[B.entry_condition_total].field = B_entry_lookup($1.other);
					strcpy(B.entry_condition[B.entry_condition_total].value,$3.value);
					B.entry_condition[B.entry_condition_total].comparison = NOT_EQUAL;
					B.entry_condition[B.entry_condition_total].type = ENTRY;
					B.entry_condition_total++;
				} 
				if (!strcmp($1.value,"CARD")) {
					if (!strcmp($1.other,"Card")) {
						B.card_condition[B.card_condition_total].field = -1;
					} else {
						B.card_condition[B.card_condition_total].field = B_card_lookup($1.other);
					}
					strcpy(B.card_condition[B.card_condition_total].value,$3.value);
					B.card_condition[B.card_condition_total].comparison = NOT_EQUAL;
					B.card_condition[B.card_condition_total].type = CARD;
					B.card_condition_total++;
				}
			}
			if ((!strcmp($1.text,"VALUE")) && (!strcmp($3.text,"B"))) {
				if (!strcmp($3.value,"ENTRY")) {
					B.entry_condition[B.entry_condition_total].field = B_entry_lookup($3.other);
					strcpy(B.entry_condition[B.entry_condition_total].value,$1.value);
					B.entry_condition[B.entry_condition_total].comparison = NOT_EQUAL;
					B.entry_condition[B.entry_condition_total].type = ENTRY;
					B.entry_condition_total++;
				} 
				if (!strcmp($3.value,"CARD")) {
					if (!strcmp($3.other,"Card")) {
						B.card_condition[B.card_condition_total].field = -1;
					} else {
						B.card_condition[B.card_condition_total].field = B_card_lookup($3.other);
					}
					strcpy(B.card_condition[B.card_condition_total].value,$1.value);
					B.card_condition[B.card_condition_total].comparison = NOT_EQUAL;
					B.card_condition[B.card_condition_total].type = CARD;
					B.card_condition_total++;
				}
			}

			if ((!strcmp($1.text,"B")) && (!strcmp($3.text,"A"))) {

				/* a,b condition -- for sort */
				if (!strcmp($1.value,"ENTRY")) {
				        B.c_condition[B.c_condition_total].field = B_entry_lookup($1.other);
					B.c_condition[B.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeA = ENTRY;
					C.conditions[C.condition_total].fieldA = B_entry_lookup($1.other);
				}
				if (!strcmp($1.value,"CARD")) {
				        B.c_condition[B.c_condition_total].field = B_card_lookup($1.other);
					B.c_condition[B.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeA = CARD;
					C.conditions[C.condition_total].fieldA = B_card_lookup($1.other);
				}
				if (!strcmp($3.value,"ENTRY")) {
				        A.c_condition[A.c_condition_total].field = A_entry_lookup($3.other);
					A.c_condition[A.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeB = ENTRY;
					C.conditions[C.condition_total].fieldB = A_entry_lookup($3.other);
				}
				if (!strcmp($3.value,"CARD")) {
				        A.c_condition[A.c_condition_total].field = A_card_lookup($3.other);
					A.c_condition[A.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeB = CARD;
					C.conditions[C.condition_total].fieldB = A_card_lookup($3.other);
				}
				A.c_condition_total++;
				B.c_condition_total++;

				/* a,b condition -- for cartesian product search */
				C.conditions[C.condition_total].comparison = NOT_EQUAL;
				C.condition_total++;

			}

			if ((!strcmp($1.text,"A")) && (!strcmp($3.text,"B"))) {
				if (!strcmp($3.value,"ENTRY")) {
				        B.c_condition[B.c_condition_total].field = B_entry_lookup($3.other);
					B.c_condition[B.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeB = ENTRY;
					C.conditions[C.condition_total].fieldB = B_entry_lookup($3.other);
				}
				if (!strcmp($3.value,"CARD")) {
				        B.c_condition[B.c_condition_total].field = B_card_lookup($3.other);
					B.c_condition[B.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeB = CARD;
					C.conditions[C.condition_total].fieldB = B_card_lookup($3.other);
				}
				if (!strcmp($1.value,"ENTRY")) {
				        A.c_condition[A.c_condition_total].field = A_entry_lookup($1.other);
					A.c_condition[A.c_condition_total].type = ENTRY;
					C.conditions[C.condition_total].typeA = ENTRY;
					C.conditions[C.condition_total].fieldA = A_entry_lookup($1.other);
				}
				if (!strcmp($1.value,"CARD")) {
				        A.c_condition[A.c_condition_total].field = A_card_lookup($1.other);
					A.c_condition[A.c_condition_total].type = CARD;
					C.conditions[C.condition_total].typeA = CARD;
					C.conditions[C.condition_total].fieldA = A_card_lookup($1.other);
				}
				A.c_condition_total++;
				B.c_condition_total++;

				/* a,b condition -- for cartesian product search */
				C.conditions[C.condition_total].comparison = NOT_EQUAL;
				C.condition_total++;
			}
			}
		}
	;

when_value: '(' NAME ')' NAME '[' when_index ']' '[' when_index ']' '[' when_index ']' 
		{
			if (!do_not_execute) {
			/* entry search */
			/* if blank is set, search all entries */
			/* if not sorted already, sort entry list for that stack by NAME */
			/* make certain stack exists. mark somewhere if sorted stack has changed. */
			if (!strcmp($2.text,"A")) {
				A.entries = 1;
				if (A.rplnum == -1) {
					A.rplnum = get_rpl_number($6.value);	
					A.rplnum_exists = 1;
				} else {
					if (A.rplnum != get_rpl_number($6.value)) {
						printf("SDL: inconsistant use of stack number for A.\n");
						printf("Line %d. Exiting.\n",file_lineno);
						exit(1);
					}
				}
				strcpy($$.text,"A");
				strcpy($$.value,"ENTRY");
				strcpy($$.other,$4.text);
			}
			if (!strcmp($2.text,"B")) {
				B.entries = 1;
				if (B.rplnum == -1) {
					B.rplnum = get_rpl_number($6.value);	
					B.rplnum_exists = 1;
				} else {
					if (B.rplnum != get_rpl_number($6.value)) {
						printf("SDL: inconsistant use of stack number for B.\n");
						printf("Line %d. Exiting.\n",file_lineno);
						exit(1);
					}
				}
				strcpy($$.text,"B");
				strcpy($$.value,"ENTRY");
				strcpy($$.other,$4.text);
			}
			}
		}
	| '(' NAME ')' NAME '[' when_index ']' '[' when_index ']'
		{
			if (!do_not_execute) {
			/* card search */
			if (!strcmp($2.text,"A")) {
				A.cards = 1;
				if (A.rplnum == -1) {
					A.rplnum = get_rpl_number($6.value);	
					A.rplnum_exists = 1;
				} else {
					if (A.rplnum != get_rpl_number($6.value)) {
						printf("SDL: inconsistant use of stack number for A.\n");
						printf("Line %d. Exiting.\n",file_lineno);
						exit(1);
					}
				}
				strcpy($$.text,"A");
				strcpy($$.value,"CARD");
				strcpy($$.other,$4.text);
			}
			if (!strcmp($2.text,"B")) {
				B.cards = 1;
				if (B.rplnum == -1) {
					B.rplnum = get_rpl_number($6.value);	
					B.rplnum_exists = 1;
				} else {
					if (B.rplnum != get_rpl_number($6.value)) {
						printf("SDL: inconsistant use of stack number for B.\n");
						printf("Line %d. Exiting.\n",file_lineno);
						exit(1);
					}
				}
				strcpy($$.text,"B");
				strcpy($$.value,"CARD");
				strcpy($$.other,$4.text);
			}
			}

		}
	| '(' NAME ')' NAME '[' when_index ']' 
		{
			if (!do_not_execute) {
			/* stack search */
			if (!strcmp($2.text,"A")) {
				if (A.rplnum == -1) {
					A.rplnum = get_rpl_number($6.value);	
					A.rplnum_exists = 1;
				} else {
					if (A.rplnum != get_rpl_number($6.value)) {
						printf("SDL: inconsistant use of stack number for A.\n");
						printf("Line %d. Exiting.\n",file_lineno);
						exit(1);
					}
				}
				strcpy($$.text,"A");
				strcpy($$.value,"STACK");
				strcpy($$.other,$4.text);
			}
			if (!strcmp($2.text,"B")) {
				if (B.rplnum == -1) {
					B.rplnum = get_rpl_number($6.value);	
					B.rplnum_exists = 1;
				} else {
					if (B.rplnum != get_rpl_number($6.value)) {
						printf("SDL: inconsistant use of stack number for B.\n");
						printf("Line %d. Exiting.\n",file_lineno);
						exit(1);
					}
				}
				strcpy($$.text,"B");
				strcpy($$.value,"STACK");
				strcpy($$.other,$4.text);
			}
			}
		} 
	| when_expr
		{
			if (!do_not_execute) {
			strcpy($$.value,$1.value);
			strcpy($$.text,"VALUE");
			}
		} 
	;

when_index: /* blank */
		{
			if (!do_not_execute) {
			strcpy($$.text,"");
			/* clear */
			}
		}
	| when_expr
	;

link_statements: link_statement
	| link_statements ',' link_statement
	;

link_statement: link_variable RPL_ASSIGN link_expr
		{
			if (!do_not_execute) {
			C.links[C.link_total].node = $3.node;
			C.links[C.link_total].list = (int) $1.text[0];
			C.links[C.link_total].type = $1.stack_int;
			if (!strcmp($1.text,"A")) {
				if ($1.stack_int == ENTRY) {
					C.links[C.link_total].field = A_entry_lookup($1.other); 
				}
				if ($1.stack_int == CARD) {
					C.links[C.link_total].field = A_card_lookup($1.other); 
				}
			}
			if (!strcmp($1.text,"B")) {
				if ($1.stack_int == ENTRY) {
					C.links[C.link_total].field = B_entry_lookup($1.other); 
				}
				if ($1.stack_int == CARD) {
					C.links[C.link_total].field = B_card_lookup($1.other); 
				}
			}
			C.link_total++;
			}
		}
	;

link_variable: '(' NAME ')' NAME '[' when_index ']' '[' when_index ']' '[' when_index ']' 
		{
			if (!do_not_execute) {
			$$.stack_int = ENTRY;
			strcpy($$.text,$2.text);
			strcpy($$.other,$4.text);
			}
		}
	| '(' NAME ')' NAME '[' when_index ']' '[' when_index ']' 
		{
			if (!do_not_execute) {
			$$.stack_int = CARD;
			strcpy($$.text,$2.text);
			strcpy($$.other,$4.text);
			}
		}
	| '(' NAME ')' NAME '[' when_index ']' 
		{
			if (!do_not_execute) {
			$$.stack_int = STACK;
			strcpy($$.text,$2.text);
			strcpy($$.other,$4.text);
			}
		}
	;

link_expr: '(' NAME ')' NAME '[' when_index ']' '[' when_index ']' '[' when_index ']' 
		{
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->list = (int) $2.text[0];
			$$.node->type = ENTRY;
			if (!strcmp($2.text,"A")) {
				$$.node->field = A_entry_lookup($4.text); 
			}
			if (!strcmp($2.text,"B")) {
				$$.node->field = B_entry_lookup($4.text); 
			}
			}
		}
	| '(' NAME ')' NAME '[' when_index ']' '[' when_index ']' 
		{
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->list = (int) $2.text[0];
			$$.node->type = CARD;
			if (!strcmp($2.text,"A")) {
				$$.node->field = A_card_lookup($4.text); 
			}
			if (!strcmp($2.text,"B")) {
				$$.node->field = B_card_lookup($4.text); 
			}
			}
		}
	| '(' NAME ')' NAME '[' when_index ']' 
		{
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->list = (int) $2.text[0];
			$$.node->type = STACK;
			$$.node->field = global_field_index($4.text); 
			}
		}
	| link_expr '-' link_expr 
		{ 
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->operator = MINUS;
			$$.node->type = OPERATOR;
			$$.node->left = $1.node;
			$$.node->right = $3.node;
			}
		}
	| link_expr '+' link_expr 
		{ 
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->operator = PLUS;
			$$.node->type = OPERATOR;
			$$.node->left = $1.node;
			$$.node->right = $3.node;
			}
		}
	| '(' link_expr ')'
		{ 
			if (!do_not_execute) {
			$$.node = $2.node;
			}
		}
	| NUMBER 		
		{ 
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->type = VALUE;
			strcpy($$.node->value,$1.value); 
			}
		}
	| STRING		
		{ 
			if (!do_not_execute) {
			$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
			$$.node->type = VALUE;
			strcpy($$.node->value,$1.value); 
			}
		}
	| NAME
		{
			/* check for badly placed names */
			if (!do_not_execute) {
				if ($1.text[0] == '$') {
					printf("SDL: source reference (using $) must have index.\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				if ($1.text[0] == '_') {
					printf("SDL: _var should only be used on left side of '<-' .\n");
					printf("Line %d. Exiting.\n",file_lineno);
					exit(1);
				}
				/* user symbol */
				$$.node = (struct LinkNode *) my_malloc(sizeof(struct LinkNode)+1);
				$$.node->type = VALUE;
				strcpy($$.node->value,get_local_value($1.text));
			}
		}
	;

given_list: data_ref
		{
			/* first one */
			if (!do_not_execute) {
				strcpy($$.value,$1.value);
			}
		}
	| given_list ',' data_ref
		{
			/* the rest */
			if (!do_not_execute) {
				strcpy($$.value,$1.value);
				strcat($$.value, " ");
				strcat($$.value, $3.value);
			}
		}
	;

data_ref: NAME '[' ']'
		{
			if (!do_not_execute) {
				strcpy($$.value,$1.text);
			}
		}
	;

book_list: given_list ',' NUMBER
		{
			if (!do_not_execute) {
				strcpy($$.value,$1.value);
				strcat($$.value, " ");
				strcat($$.value, $3.value);
			}
		}
	| given_list
		{
			if (!do_not_execute) {
				strcpy($$.value,$1.value);
				strcat($$.value, " ");
				strcat($$.value, "-1");
			}
		}
	;

%%

just_for_debugging()
{
}

unsigned char *my_malloc(size)
int size;
{
	/* unsigned char *temp; */

#				ifdef SD_DEBUG3
				printf("S:MM");
				printf("%d\n",mem_tot++);
				printf("size %d\n",size);
#				endif
	if (size < 1)
	{
		printf("SD_RPL: Attempt to allocate less than one byte.\n");
		printf("%d Exiting.\n",size);
		exit(1);
	}
#				ifdef SD_DEBUG3
	printf("now\n");
	if (mem_tot == 444473) 
	{
		just_for_debugging();
	}
#				endif
	mem_temp = (unsigned char *)calloc(1,size);
#				ifdef SD_DEBUG3
				printf("1temp:%d\n",mem_temp);
				printf("\n\n\n");
#				endif
	if (mem_temp == (unsigned char *) NULL) {
		printf("SDL: Out of memory. Exiting.\n");
		exit(1);
	}
#				ifdef SD_DEBUG3
				printf("2temp:%d\n",mem_temp);
				printf("\n\n\n");
#				endif
	memset(mem_temp,0,size);
#				ifdef SD_DEBUG3
				printf("F:MM\n");
#				endif
	return mem_temp;
}

my_free(ptr)
unsigned char *ptr;
{
	if (ptr == (unsigned char *)0) 
	{
		printf("SD_RPL: Attempt to free a zero ptr.\n");
		printf("Exiting.\n");
		exit(1);
	}
	free(ptr);
}

#include "sdl_l.c"

int sdl_flush_lex_buffers()
{
  YY_FLUSH_BUFFER;
}

	/*               storage < ----- | ----- > values                          */

page(ts1,ts2,ss)
unsigned char *ts1;
unsigned char *ts2;
unsigned char *ss;
{
	int i,j,k;
	for (k=0; ss[k]; k++) {
		if (isdigit(ss[k]))
			ts1[k] = ss[k];
		else break;
	}
	ts1[k++] = '\0';
	for (j=0;ss[k];j++,k++) {
		if (isdigit(ss[k])) 
			ts2[j] = ss[k];
		else break;
	}
	ts2[k] = '\0';
}

int data_file_size(fd)
FILE *fd;
{
	/* assume newly opened fd */
	int i,size;
	unsigned char buf[500];

	data_lines = 0;
	size = 0;
	for (i=0;fgets(buf,400,fd);i++) {
		size += strlen(buf);
		data_lines++;
	}
	fseek(fd,0,0);
	return size;
	/* ... set records count in read_data. make user-accesible. */
}

int card_field_index(field_name,card_type)
unsigned char *field_name;
unsigned char *card_type;
{
	int c_index, f_index;

	if (!strcmp(field_name,"Card")) return 0;
	for (c_index = 0;strlen(card_types[c_index].name);c_index++) 
		if (!strcmp(card_types[c_index].name,card_type)) break;
	if (!strlen(card_types[c_index].name)) {
		printf("SDL: Couldn't find card type %s. Line %d. Exiting.\n",
			card_type,file_lineno);
		exit(1);
	}
	for (f_index = 0;strlen(card_types[c_index].field_names[f_index]);f_index++) 
		if (!strcmp(card_types[c_index].field_names[f_index],field_name)) break;
	if (!strlen(card_types[c_index].field_names[f_index])) {
		printf("SDL: Couldn't find field_name %s in card type %s. Line %d. Exiting.\n",
			field_name,card_type,file_lineno);
		exit(1);
	}
	return f_index;
}

int entry_field_index(field_name,card_type)
unsigned char *field_name;
unsigned char *card_type;
{
	int c_index, f_index;

	for (c_index = 0;strlen(card_types[c_index].name);c_index++) 
		if (!strcmp(card_types[c_index].name,card_type)) break;
	if (!strlen(card_types[c_index].name)) {
		printf("SDL: Couldn't find card type %s. Line %d. Exiting.\n",
			card_type,file_lineno);
		exit(1);
	}
	for (f_index = 0;strlen(card_types[c_index].list_field_names[f_index]);f_index++) 
		if (!strcmp(card_types[c_index].list_field_names[f_index],field_name)) break;
	if (!strlen(card_types[c_index].list_field_names[f_index])) {
		printf("SDL: Couldn't find list_field_name %s in card type %s. Line %d. Exiting.\n",
			field_name,card_type,file_lineno);
		exit(1);
	}
	return f_index;
}

int get_rpl_number(stack_number)
unsigned char *stack_number;
{
	int index, i;

	index = global_field_index("StackNum");
	for (i=0;i<rpl_index;i++) {
		if (!rpl_list[i]->fields[index]) continue;	
		if (!strcmp(stack_number,rpl_list[i]->fields[index])) return i;	
	}
	printf("SDL: could not find requested stack number among rpls.\n");
	printf("Line %d. Exiting.\n",file_lineno);
	exit(1);
}

struct Card *get_card(rpl_number,card_index)
int rpl_number;
int card_index;
{
	struct Card *temp_card;

	temp_card = rpl_list[rpl_number]->cards[card_index];
	if (!temp_card) {
		printf("SDL: Couldn't find CARD %d in rpl stack %d. Line %d. Exiting.\n",
			card_index,rpl_number,file_lineno);
		exit(1);
	}
	return temp_card;
}

struct Entry *get_entry(card,entry_index)
struct Card *card;
int entry_index;
{
	struct Entry *temp_entry;
	
	temp_entry = card->entries[entry_index];
	if (!temp_entry){
		printf("SDL: Couldn't find entry %d given card ptr %d. Line %d. Exiting.\n",
			entry_index,card,file_lineno);
		exit(1);
	}
	return temp_entry;
}

extern unsigned char country[];

read_sdl()
{
	int i;

	assign_stack = 0;
	assign_card = (struct Card *) 0;
	assign_entry = (struct Entry *) 0;
	assign_name = 0;
	in_new_stack = 0;
	in_new_card = 0;
	in_new_entry = 0;
	do_not_execute = 0;
	A.rplnum = -1;
	B.rplnum = -1;
#ifdef SD_DEBUG4
	printf("%5d ",1);
#endif
	/* zero A and B */
	memset(&A,0,sizeof(A));
	memset(&B,0,sizeof(B));
	memset(&C,0,sizeof(C));

	ptl_index = 0;
	memset(page_total_list,0,sizeof(page_total_list));
	card_book_count = 1;

	A.rplnum = -1;
	B.rplnum = -1;

//	sdl_debug = 1;
	// sdl_restart(sdl_fd);
	sdl_parse();

}

int data_field_index(field_name)
unsigned char *field_name;
{
	int i;

	for (i=0; strlen(field_names[i]); i++) 
		if (!strcmp(field_names[i],field_name))
			return i;
	printf("Could not find field name %s. Line %d. Exiting. \n",
		field_name,file_lineno);
	exit(1);
}

int global_field_index(s)
unsigned char *s;
{
	int g_index;

	for (g_index=0;strlen(global_defs[g_index]);g_index++) 
		if (!strcmp(global_defs[g_index],s)) return g_index;
	printf("SDL: could not find global %s. Line %d. Exiting.\n",s,file_lineno);
	exit(1);
}

#define SYMTABSIZE (200)
struct pairs {
	int used;
	unsigned char name[50];
	unsigned char value[100];
} symbol_table[SYMTABSIZE];
int symbol_table_index = 0;

int user_exist(s)
unsigned char *s;
{
	int i;

	for (i=0; i<symbol_table_index; i++) 
		if (symbol_table[i].used && !strcmp(symbol_table[i].name,s))
			return 1;
	return 0;
}

unsigned char *get_local_value(s)
unsigned char *s;
{
	int i;

	for (i=0; i<symbol_table_index; i++) 
		if (symbol_table[i].used && !strcmp(symbol_table[i].name,s))
			return symbol_table[i].value;
	printf("SDL: user variable %s accessed before value assigned. Line %d. Exiting. \n",s,file_lineno);
	exit(1);	
}

set_local_value(s1,s2)
unsigned char *s1;
unsigned char *s2;
{
	int i;

	for (i=0; i<symbol_table_index; i++) {
		if (symbol_table[i].used && !strcmp(symbol_table[i].name,s1)) {
			strcpy(symbol_table[i].value,s2);
			return;
		}
	}
	strcpy(symbol_table[i].name,s1);
	strcpy(symbol_table[i].value,s2);
	symbol_table[i].used = 1;
	symbol_table_index++;
	if (symbol_table_index > SYMTABSIZE) {
	  printf("\nSymbol table overflow on line %d: %s Exiting\n",file_lineno,yytext);
	  exit(1);
	}
}

yyerror(char *s)
{
	printf("\nSyntax error on line %d: %s Exiting\n",file_lineno,yytext);
	printf("(yacc message -- %s)\n",s);
	exit(1);
}

unsigned char *data_field(index,field)
unsigned int index;
unsigned int field;
{
	int i,j,k;
	int count = 0;
	unsigned char buf[1000];

#if 1
	if (index <= data_records) {
	  unsigned char *m;
	  unsigned char *s = data_ptr[index];
	  while (1) {
	    if (count == field) {
	      unsigned char *b = s;
	      while (*s && *s != '|') {
		s++;
	      }
	      m = my_malloc(s-b+1);
	      strncpy(m,b,s-b);
	      m[s-b] = 0;
	      return m;
	    }
	    while (*s && *s != '|') {
	      s++;
	    }
	    if (!*s) {
#if 0
	      printf("SDL: Couldn't get data. Line %d. Exiting. \n",file_lineno);
	      exit(1);
#else
		return (unsigned char *) strcpy(my_malloc(3),"");
#endif
	    }
	    s++;
	    count++;
	  }
	  
	}

#else
	for (i=0;data_ptr[index][i];i++) {
		if (count == field) {
			for (j=i,k=0;(data_ptr[index][j] != '|') && (data_ptr[index][j]);
				j++,k++) { 
			  if (k>1000) {
			    printf("SDL: copy too big. Line %d. Exiting. \n",file_lineno);
			    exit(1);
			  }
				buf[k] = data_ptr[index][j];
			}
			buf[k] = '\0';
			return (unsigned char *) strcpy(my_malloc(k+2),buf);
		}
		if (data_ptr[index][i] == '|') {
			count++;
		}
	}
#endif
#if 0	
	printf("SDL: Couldn't get data. Line %d. Exiting. \n",file_lineno);
	exit(1);
#else
	return (unsigned char *) strcpy(my_malloc(3),"");
#endif
}

erase_keys()
{
	int i;

	for (i=0;i<number_of_keys;i++) {
		key_type[i] = 0;
		key_field[i][0] = '\0';
	}
	number_of_keys = 0;
}

#define SORT_TYPES 8

struct {
	int code;
	unsigned char *string;
} sort_type[] = { ALPHA, (unsigned char *) "ALPHA",
		REV_ALPHA , (unsigned char *) "REV_ALPHA",
		FORK , (unsigned char *) "FORK",
		PAV , (unsigned char *) "PAV",
		DATE , (unsigned char *) "DATE",
		FLEUR, (unsigned char *) "FLEUR", 
		NUMERICAL, (unsigned char *) "NUMERICAL",
		ALPHANUMERIC, (unsigned char *) "ALPHANUMERIC" };

int string_to_key(s)
unsigned char *s;
{
	int i;
	for (i=0;i<SORT_TYPES; i++) {
		if (!strcmp(sort_type[i].string,s))
			return sort_type[i].code;
	}
	printf("SDL: Unknown sort type %s. Line %d. Exiting.\n",s,file_lineno);
	exit(1);
}

int rev_alpha_compare(s1,s2)
unsigned char *s1;
unsigned char *s2;
{
	int result;

	result = strcmp(s1,s2);
	if (!result) return 0;
	if (result < 0) {
		/* s1 < s2 -- out of order */
		return 2;
	} else {
		/* s1 > s2 */
		return 1;
	}
}

int alpha_compare(s1,s2)
unsigned char *s1;
unsigned char *s2;
{
	int result;

	result = strcmp(s1,s2);
	if (!result) return 0;
	if (result < 0) {
		/* s1 < s2 */
		return 1;
	} else {
		/* s1 > s2 -- out of order */
		return 2;
	}
}

unsigned char *forks[] = {
	(unsigned char *) "F58D",
	(unsigned char *) "F596",
	(unsigned char *) "F58C",
	(unsigned char *) "F595",
	(unsigned char *) "F58B",
	(unsigned char *) "F594",
	(unsigned char *) "F58A",
	(unsigned char *) "F241",
	(unsigned char *) "F589",
	(unsigned char *) "F371",
	(unsigned char *) "F0BF",
	(unsigned char *) ""
	};

int fork_compare(s1,s2)
unsigned char *s1;
unsigned char *s2;
{
	int rank1, rank2, i;

	for (i=0; i< 10; i++) {
		if (!strcmp(s1,forks[i]))
			break;
	}
	rank1 = i;
	for (i=0; i< 10; i++) {
		if (!strcmp(s2,forks[i]))
			break;
	}
	rank2 = i;

	if (rank1 == rank2) return 0;
	if (rank1 < rank2) {
		/* s1 < s2 */
		return 1;
	} else {
		/* s1 > s2 -- out of order */
		return 2;
	}
}


unsigned char *pavs[] = {
	(unsigned char *) "F592",
	(unsigned char *) "F580",
	(unsigned char *) "F591",
	(unsigned char *) "F583",
	(unsigned char *) "F590",
	(unsigned char *) "F5A8",
	(unsigned char *) "F58F",
	(unsigned char *) "F39C",
	(unsigned char *) "F58E",
	(unsigned char *) "F38E",
	(unsigned char *) "F0CC",
	(unsigned char *) ""
	};

int pav_compare(s1,s2)
unsigned char *s1;
unsigned char *s2;
{
	int rank1, rank2, i;

	for (i=0; i< 11; i++) {
		if (!strcmp(s1,pavs[i]))
			break;
	}
	rank1 = i;
	for (i=0; i< 11; i++) {
		if (!strcmp(s2,pavs[i]))
			break;
	}
	rank2 = i;

	if (rank1 == rank2) return 0;
	if (rank1 < rank2) {
		/* s1 < s2 */
		return 1;
	} else {
		/* s1 > s2 -- out of order */
		return 2;
	}
}

unsigned char *date[] = {
	(unsigned char *) "Januar/Februar",
	(unsigned char *) "Mrz/April",
	(unsigned char *) "Mai/Juni",
	(unsigned char *) "Juli/August",
	(unsigned char *) "September/Oktober",
	(unsigned char *) "November/Dezember",
	(unsigned char *) "January",
	(unsigned char *) "February",
	(unsigned char *) "February-March",
	(unsigned char *) "March",
	(unsigned char *) "April",
	(unsigned char *) "May",
	(unsigned char *) "Spring",
	(unsigned char *) "June",
	(unsigned char *) "July",
	(unsigned char *) "July-August",
	(unsigned char *) "August",
	(unsigned char *) "Summer",
	(unsigned char *) "September",
	(unsigned char *) "October",
	(unsigned char *) "November",
	(unsigned char *) "Fall",
	(unsigned char *) "December",
	(unsigned char *) "Winter",
	(unsigned char *) "Year-Round",
	(unsigned char *) "Olympic Events",
	(unsigned char *) "Ongoing",
	(unsigned char *) ""
	};

unsigned char *fleurs[] = {
	(unsigned char *) "F598",
	(unsigned char *) "F597",
	(unsigned char *) "F0A6",
	(unsigned char *) "F0CD",
	(unsigned char *) ""
	};

int fleur_compare(s1,s2)
unsigned char *s1;
unsigned char *s2;
{
	int rank1, rank2, i;

	for (i=0; i< 3; i++) {
		if (!strcmp(s1,fleurs[i]))
			break;
	}
	rank1 = i;
	for (i=0; i< 3; i++) {
		if (!strcmp(s2,fleurs[i]))
			break;
	}
	rank2 = i;

	if (rank1 == rank2) return 0;
	if (rank1 < rank2) {
		/* s1 < s2 */
		return 1;
	} else {
		/* s1 > s2 -- out of order */
		return 2;
	}
}

int
qcompare(unsigned char** a,unsigned char** b)
{
  int index1 = ((a-data_ptr));
  int index2 = ((b-data_ptr));
	int cr;
	int field_index;
	int rank1, rank2, i;
 	for (i=0;i<number_of_keys;i++) {
		field_index = data_field_index(key_field[i]);
		switch(key_type[i]) {
		case ALPHANUMERIC:
		  {
		    unsigned char *s1,*s2;
		    s1= data_field(index1,field_index);
		    s2= data_field(index2,field_index);
		    cr = strcmp(s1,s2);
		    my_free(s1);my_free(s2);
		    switch(cr) {
		    case 0:	/* same */
		      break;
		    default:
		      return cr;
		    }
		  }
		  break;
		case NUMERICAL:
		  {
		    unsigned char *s1,*s2;
		    s1= data_field(index1,field_index);
		    s2= data_field(index2,field_index);
		    cr = atoi(s1) - atoi(s2);
		    /*cr = num_compare(s1,s2);*/
		    my_free(s1);my_free(s2);
		    switch(cr) {
		    case 0:	/* same */
		      break;
		    default:
		      return cr;
		    }
		  }
		  break;
		case ALPHA:
		  {
		    unsigned char *s1,*s2;
		    s1= data_field(index1,field_index);
		    s2= data_field(index2,field_index);
		    cr = simple_compare(s1,s2);
		    my_free(s1);my_free(s2);
		    switch(cr) {
		    case 0:	/* same */
		      break;
		    default:
		      return cr;
		    }
		  }
		  break;
		case REV_ALPHA:
		  {
		    unsigned char *s1,*s2;
		    s1= data_field(index1,field_index);
		    s2= data_field(index2,field_index);
		    cr = strcasecmp(s1,s2);
		    my_free(s1);my_free(s2);
		    switch(cr) {
		    case 0:	/* same */
		      break;
		    default:
		      return -cr;
		    }
		  }
		  break;
		case FORK:
		  {
		    unsigned char *s1= data_field(index1,field_index);
		    unsigned char *s2= data_field(index2,field_index);
		    int rank1, rank2, i;

#define NFORKS (sizeof(forks)/sizeof(forks[0]))
		    for (i=0; i< NFORKS; i++) {
		      if (!strcasecmp(s1,forks[i]))
			break;
		    }
		    if (i>=NFORKS) {
		      printf("whoops forks %s!\n",s1);
		      exit(1);
		    }
		    rank1 = i;
		    for (i=0; i< NFORKS; i++) {
		      if (!strcasecmp(s2,forks[i]))
			break;
		    }
		    if (i>=NFORKS) {
		      printf("whoops forks %s!\n",s2);
		      exit(1);
		    }
		    rank2 = i;
		    my_free(s1);my_free(s2);

		    if (rank1 == rank2) break;
		    if (rank1 < rank2) {
		      /* s1 < s2 */
		      return -1;
		    } else {
		      /* s1 > s2 */
		      return 1;
		    }
		  }
		  break;
		case PAV:
		  {
		    unsigned char *s1= data_field(index1,field_index);
		    unsigned char *s2= data_field(index2,field_index);
		    int rank1, rank2, i;

#define NPAVS (sizeof(pavs)/sizeof(pavs[0]))
		    for (i=0; i< NPAVS; i++) {
		      if (!strcasecmp(s1,pavs[i]))
			break;
		    }
		    if (i>=NPAVS) {
		      printf("whoops pavs %s!\n",s1);
		      exit(1);
		    }
		    rank1 = i;
		    for (i=0; i< NPAVS; i++) {
		      if (!strcasecmp(s2,pavs[i]))
			break;
		    }
		    if (i>=NPAVS) {
		      printf("whoops pavs %s!\n",s2);
		      exit(1);
		    }
		    rank2 = i;
		    my_free(s1);my_free(s2);

		    if (rank1 == rank2) break;
		    if (rank1 < rank2) {
		      /* s1 < s2 */
		      return -1;
		    } else {
		      /* s1 > s2 */
		      return 1;
		    }
		  }
		  break;
		case DATE:
		  {
		    unsigned char *s1= data_field(index1,field_index);
		    unsigned char *s2= data_field(index2,field_index);
		    int rank1, rank2, i;

#define NDATE (sizeof(date)/sizeof(date[0]))
		    for (i=0; i< NDATE; i++) {
		      if (!strcasecmp(s1,date[i]))
			break;
		    }
		    if (i>=NDATE) {
		      printf("whoops DATE %s!\n",s1);
		      exit(1);
		    }
		    rank1 = i;
		    for (i=0; i< NDATE; i++) {
		      if (!strcasecmp(s2,date[i]))
			break;
		    }
		    if (i>=NDATE) {
		      printf("whoops date %s!\n",s2);
		      exit(1);
		    }
		    rank2 = i;
		    my_free(s1);my_free(s2);

		    if (rank1 == rank2) break;
		    if (rank1 < rank2) {
		      /* s1 < s2 */
		      return -1;
		    } else {
		      /* s1 > s2 */
		      return 1;
		    }
		  }
		  break;
		case FLEUR:
		  {
		    unsigned char *s1= data_field(index1,field_index);
		    unsigned char *s2= data_field(index2,field_index);
		    int rank1, rank2, i;

#define NFLEURS (sizeof(fleurs)/sizeof(fleurs[0]))
		    for (i=0; i< NFLEURS; i++) {
		      if (!strcasecmp(s1,fleurs[i]))
			break;
		    }
		    if (i>=NFLEURS) {
		      printf("whoops pfleurs %s!\n",s1);
		      exit(1);
		    }
		    rank1 = i;
		    for (i=0; i< NFLEURS; i++) {
		      if (!strcasecmp(s2,fleurs[i]))
			break;
		    }
		    if (i>=NFLEURS) {
		      printf("whoops pfleurs %s!\n",s2);
		      exit(1);
		    }
		    rank2 = i;
		    my_free(s1);my_free(s2);

		    if (rank1 == rank2) break;
		    if (rank1 < rank2) {
		      /* s1 < s2 */
		      return -1;
		    } else {
		      /* s1 > s2 */
		      return 1;
		    }
		  }
		  break;
		default:
		  printf("whoops bad key type %d\n",key_type[i]);
		  exit(1);
		  break;

		}

	}
  return 0;
}

sort()
{
	qsort(data_ptr,data_records-1,sizeof(unsigned char*),(int (*)(const void *, const void *))&qcompare);
}

unsigned char *itoa(n,s)
int n;
unsigned char *s;
{
	int i, sign;

	if ((sign = n) < 0)
		n = -n;
	i=0;
	do {
		s[i++] = n % 10 + '0';
	} while ((n/= 10) > 0);
	if (sign < 0)
		s[i++] = '-';
	s[i] = '\0';
	reverse(s);
	return s;
}

reverse(s)
unsigned char *s;
{
	int c,i,j;

	for (i=0, j = strlen(s)-1; i<j;i++,j--){
		c = s[i];
		s[i] = s[j];
		s[j] = c;
	}
}

int myindex(source,search)
unsigned char *source;
unsigned char *search;
{
	int i;
	for (i=0;source[i];i++){
		if (!strcmp(&source[i],search)) {
			return i;
		}
	}
	return 0;
}

/* (if we don't have any sorted sets that match,) 
(1) we create lists a and/or b restricted by conditions (a,b).
The lists are of entry, card or stack pointers 
depending on the set references in (c). */

/*int qcompare(const void *, const void *);
int A_entry_compare(const void *, const void *);
int A_card_compare(const void *, const void *);
int B_entry_compare(const void *, const void *);
int B_card_compare(const void *, const void *);*/
int A_entry_compare(struct Entry** entry1, struct Entry** entry2);
int A_card_compare(struct Card ** card1, struct Card** card2);
int B_entry_compare(struct Entry** entry1, struct Entry** entry2);
int B_card_compare(struct Card ** card1, struct Card** card2);
void a_b_compare();
void do_links(int a, int b);
unsigned char *link_eval( struct LinkNode * node, int a, int b);
int card_eval(struct Condition *condition, struct Card *card);
int entry_eval(struct Condition *condition, struct Entry *entry);
void clean_a_b();

void create_a()
{
	int a,b,c,d;

	if (A.entries) {
		if (!A.rplnum_exists) {
			printf("When-link error: currently must provide stack number.\n");
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		/* initialize list */
		A.entry_list = (struct Entry **) my_malloc(ENTRY_PTR_BLOCK*sizeof(struct Entry *));
		A.list_index = 0;
		/* find or make list of entry pointers for that stack. */
		for (a=1;a<rpl_list[A.rplnum]->card_index;a++) {
			for (b=0; b<A.card_condition_total; b++) {
				if (!card_eval(&A.card_condition[b],rpl_list[A.rplnum]->cards[a])) {
					break;
				}
			}
			if (b!=A.card_condition_total){
				continue;
			}
			for (c=0; rpl_list[A.rplnum]->cards[a]->entries[c] && rpl_list[A.rplnum]->cards[a]->entries[c]->used; c++) {
				for (d=0; d<A.entry_condition_total; d++) {
					if (!entry_eval(&A.entry_condition[d],rpl_list[A.rplnum]->
						cards[a]->entries[c]))
						break;
				}
				if (d != A.entry_condition_total) {
					continue;
				}
				/* All hoops jumped: we have an entry. */
				A.entry_list[A.list_index] = rpl_list[A.rplnum]->
					cards[a]->entries[c];
				A.list_index++;
				if (!(A.list_index%ENTRY_PTR_BLOCK)) {
					int blocks, size;
					blocks = A.list_index/ENTRY_PTR_BLOCK;
					size = (blocks+1)*ENTRY_PTR_BLOCK*(sizeof(struct Entry *));
					if ((A.entry_list = (struct Entry **) realloc(A.entry_list,size))==NULL) {
						printf("REALLOC FAILS. Exiting.\n");
						exit(1);
					}
					memset(&A.entry_list[A.list_index],0,ENTRY_PTR_BLOCK*(sizeof(struct Entry *)));
				}
			}
		}
	} else {
		if (!A.cards) {
			printf("When-link error: no card or entry conditions found.\n");
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		if (!A.rplnum_exists) {
			printf("When-link error: currently must provide stack number.\n");
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		/* initialize list */
		A.card_list = (struct Card **) my_malloc(CARD_PTR_BLOCK*sizeof(struct Card *));
		A.list_index = 0;
		/* find or make list of entry pointers for that stack. */
		for (a=1;a<rpl_list[A.rplnum]->card_index;a++) {
			for (b=0; b<A.card_condition_total; b++) {
				if (!card_eval(&A.card_condition[b],rpl_list[A.rplnum]->cards[a])) {
					break;
				}
			}
			if (b!=A.card_condition_total){
				continue;
			}
			/* We have a card. */
			A.card_list[A.list_index] = rpl_list[A.rplnum]->cards[a];
			A.list_index++;
			if (!(A.list_index%CARD_PTR_BLOCK)) {
				int blocks, size;
				blocks = A.list_index/CARD_PTR_BLOCK;
				size = (blocks+1)*CARD_PTR_BLOCK*(sizeof(struct Card *));
				if ((A.card_list = (struct Card **) realloc(A.card_list,size))==NULL) {
					printf("REALLOC FAILS. Exiting.\n");
					exit(1);
				}
				memset(&A.card_list[A.list_index],0,CARD_PTR_BLOCK*(sizeof(struct Card *)));
			}
		}
	}
}

void create_b()
{
	int a,b,c,d;

	if (B.entries) {
		if (!B.rplnum_exists) {
			printf("When-link error: currently must provide stack number.\n");
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		/* initialize list */
		B.entry_list = (struct Entry **) my_malloc(ENTRY_PTR_BLOCK*sizeof(struct Entry *));
		B.list_index = 0;
		/* find or make list of entry pointers for that stack. */
		for (a=1;a<rpl_list[B.rplnum]->card_index;a++) {
			for (b=0; b<B.card_condition_total; b++) {
				if (!card_eval(&B.card_condition[b],rpl_list[B.rplnum]->cards[a])) {
					break;
				}
			}
			if (b!=B.card_condition_total){
				continue;
			}
			for (c=0; rpl_list[B.rplnum]->cards[a]->entries[c] && rpl_list[B.rplnum]->cards[a]->entries[c]->used; c++) { 
				for (d=0; d<B.entry_condition_total; d++) {
					if (!entry_eval(&B.entry_condition[d],rpl_list[B.rplnum]->
						cards[a]->entries[c]))
						break;
				}
				if (d != B.entry_condition_total) {
					continue;
				}
				/* All hoops jumped: we have an entry. */
				B.entry_list[B.list_index] = rpl_list[B.rplnum]->
					cards[a]->entries[c];
				B.list_index++;
				if (!(B.list_index%ENTRY_PTR_BLOCK)) {
					int blocks, size;
					blocks = B.list_index/ENTRY_PTR_BLOCK;
					size = (blocks+1)*ENTRY_PTR_BLOCK*(sizeof(struct Entry *));
					if ((B.entry_list = (struct Entry **) realloc(B.entry_list,size))==NULL) {
						printf("REALLOC FAILS. Exiting.\n");
						exit(1);
					}
					memset(&B.entry_list[B.list_index],0,ENTRY_PTR_BLOCK*(sizeof(struct Entry *)));
				}
			}
		}
	} else {
		if (!B.cards) {
			printf("When-link error: no card or entry conditions found.\n");
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		if (!B.rplnum_exists) {
			printf("When-link error: currently must provide stack number.\n");
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		/* initialize list */
		B.card_list = (struct Card **) my_malloc(CARD_PTR_BLOCK*sizeof(struct Card *));
		B.list_index = 0;
		/* find or make list of entry pointers for that stack. */
		for (a=1;a<rpl_list[B.rplnum]->card_index;a++) {
			for (b=0; b<B.card_condition_total; b++) {
				if (!card_eval(&B.card_condition[b],rpl_list[B.rplnum]->cards[a])) {
					break;
				}
			}
			if (b!=B.card_condition_total){
				continue;
			}
			/* We have a card. */
			B.card_list[B.list_index] = rpl_list[B.rplnum]->cards[a];
			B.list_index++;
			if (!(B.list_index%CARD_PTR_BLOCK)) {
				int blocks, size;
				blocks = B.list_index/CARD_PTR_BLOCK;
				size = (blocks+1)*CARD_PTR_BLOCK*(sizeof(struct Card *));
				if ((B.card_list = (struct Card **) realloc(B.card_list,size))==NULL) {
					printf("REALLOC FAILS. Exiting.\n");
					exit(1);
				}
				memset(&B.card_list[B.list_index],0,CARD_PTR_BLOCK*(sizeof(struct Card *)));
			}
		}
	}
}

/* Sort these sets by ALL keys in (c) conditions. 
Make sure that (whether borrowing a sort or
making a new one) the sorting keys are
in exactly the same order for a and b:
the order found. */

void sort_a()
{
	if (A.entries) {
		qsort(A.entry_list,A.list_index,sizeof(struct Entry *),(int (*)(const void *, const void *))&A_entry_compare);
	} else {
		qsort(A.card_list,A.list_index,sizeof(struct Card *),(int (*)(const void *, const void *))&A_card_compare);
	}
}

void sort_b()
{
	if (B.entries) {
		qsort(B.entry_list,B.list_index,sizeof(struct Entry *),(int (*)(const void *, const void *))&B_entry_compare);
	} else {
		qsort(B.card_list,B.list_index,sizeof(struct Card *),(int (*)(const void *, const void *))&B_card_compare);
	}
}

int A_entry_compare(struct Entry** entry1, struct Entry** entry2)
{
	int a,val;

	for (a=0; a<A.c_condition_total; a++) {
		if (A.c_condition[a].type == ENTRY) {
			if (!(*entry1)->fields[A.c_condition[a].field] ) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified A entry field.\n");
				printf("field number %d, card type %s\n",
					A.c_condition[a].field,
					(*entry1)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (!(*entry2)->fields[A.c_condition[a].field]) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified A entry field.\n");
				printf("field number %d, card type %s\n",
					A.c_condition[a].field,
					(*entry2)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (val = strcmp((*entry1)->fields[A.c_condition[a].field],
				(*entry2)->fields[A.c_condition[a].field])) {
				return val;
			}
			continue;
		}
		if (A.c_condition[a].type == CARD) {
			if (!(*entry1)->card->fields[A.c_condition[a].field] ) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified A Card field.\n");
				printf("field number %d, card type %s\n",
					A.c_condition[a].field,
					(*entry1)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (!(*entry2)->card->fields[A.c_condition[a].field]) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified A Card field.\n");
				printf("Line: %d. Exiting.\n",file_lineno);
				printf("field number %d, card type %s\n",
					A.c_condition[a].field,
					(*entry2)->card->name);
				exit(1);
			}
			if (val = strcmp((*entry1)->card->fields[A.c_condition[a].field],
				(*entry2)->card->fields[A.c_condition[a].field])) {
				return val;
			}
		}
	}
	return 0;
}

int A_card_compare(struct Card ** card1, struct Card** card2)
{
	int a,val;

	for (a=0; a<A.c_condition_total; a++) {
		if (!(*card1)->fields[A.c_condition[a].field]) {
			printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
			printf("There is no data in the specified A Card field.\n");
			printf("field number %d, card type %s\n",
				A.c_condition[a].field,
				(*card1)->name);
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		if (!(*card2)->fields[A.c_condition[a].field]) {
			printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
			printf("There is no data in the specified A Card field.\n");
			printf("field number %d, card type %s\n",
				A.c_condition[a].field,
				(*card2)->name);
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		if (val = strcmp((*card1)->fields[A.c_condition[a].field],
			(*card2)->fields[A.c_condition[a].field])) {
			return val;
		}
	}
	return 0;
}

int B_entry_compare(struct Entry** entry1, struct Entry** entry2)
{
	int a,val;

	for (a=0; a<B.c_condition_total; a++) {
		if (B.c_condition[a].type == ENTRY) {
			if (!(*entry1)->fields[B.c_condition[a].field]) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified B entry field.\n");
				printf("field number %d, card type %s\n",
					B.c_condition[a].field,
					(*entry1)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (!(*entry2)->fields[B.c_condition[a].field]) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified B entry field.\n");
				printf("field number %d, card type %s\n",
					B.c_condition[a].field,
					(*entry2)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (val = strcmp((*entry1)->fields[B.c_condition[a].field],
				(*entry2)->fields[B.c_condition[a].field])) {
				return val;
			}
			continue;
		}
		if (B.c_condition[a].type == CARD) {
			if (!(*entry1)->card->fields[B.c_condition[a].field]) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified B card field.\n");
				printf("field number %d, card type %s\n",
					B.c_condition[a].field,
					(*entry1)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (!(*entry2)->card->fields[B.c_condition[a].field]) {
				printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
				printf("There is no data in the specified B card field.\n");
				printf("field number %d, card type %s\n",
					B.c_condition[a].field,
					(*entry2)->card->name);
				printf("Line: %d. Exiting.\n",file_lineno);
				exit(1);
			}
			if (val = strcmp((*entry1)->card->fields[B.c_condition[a].field],
				(*entry2)->card->fields[B.c_condition[a].field])) {
				return val;
			}
		}
	}
	return 0;
}

int B_card_compare(struct Card ** card1, struct Card** card2)
{
	int a,val;

	for (a=0; a<B.c_condition_total; a++) {
		if (!(*card1)->fields[B.c_condition[a].field]) {
			printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
			printf("There is no data in the specified B card field.\n");
			printf("field number %d, card type %s\n",
				B.c_condition[a].field,
				(*card1)->name);
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		if (!(*card2)->fields[B.c_condition[a].field]) {
			printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
			printf("There is no data in the specified B card field.\n");
			printf("field number %d, card type %s\n",
				B.c_condition[a].field,
				(*card2)->name);
			printf("Line: %d. Exiting.\n",file_lineno);
			exit(1);
		}
		if (val = strcmp((*card1)->fields[B.c_condition[a].field],
			(*card2)->fields[B.c_condition[a].field])) {
			return val;
		}
	}
	return 0;
}

/* after the sorting is taken care of, the search begins. 
It's the same loop for all cases. Simply execute the
appropriate assignments with the appropriate values.
Simply traverse the pre-built condition tree. */

void a_b_compare()
{
	int i,j, k, a, b, val, skip;
	unsigned char *string1, *string2, *error_name;

	/* !!!Hopefully all the 'equals' comparisons are stored first, since they determine the range
		of an efficient search, whereas 'not equals' only eliminates
		a single found case. */
	k=0;
	for (i=0;i<A.list_index;i++) {
		for (j=k;j<B.list_index;j++) {
			skip = 0;
			for (a=0; a<C.condition_total && !skip;a++) {
				if (A.entries) {
					if (C.conditions[a].typeA == ENTRY) {
						string1 = A.entry_list[i]->fields[C.conditions[a].fieldA];
						error_name = A.entry_list[i]->card->name;
					}
					else {
						string1 = A.entry_list[i]->card->fields[C.conditions[a].fieldA];
						error_name = A.entry_list[i]->card->name;
					}
				} else {
					string1 = A.card_list[i]->fields[C.conditions[a].fieldA];
					error_name = A.card_list[i]->name;
				}

				if (B.entries) {
					if (C.conditions[a].typeB == ENTRY)
						string2 = B.entry_list[j]->fields[C.conditions[a].fieldB];
					else
						string2 = B.entry_list[j]->card->fields[C.conditions[a].fieldB];
					error_name = B.entry_list[j]->card->name;
				} else {
					string2 = B.card_list[j]->fields[C.conditions[a].fieldB];
					error_name = B.card_list[j]->name;
				}
				if (string1== NULL) {
			                             printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
			                             printf("There is no data in the specified A card field.\n");
			                             printf("field number %d, card type %s\n",
				                             C.conditions[a].fieldA,
				                             error_name);
			                             printf("Line: %d. Exiting.\n",file_lineno);
			                             exit(1);
		                }
						    
				if (string2 == NULL) {
			                             printf("FATAL ERROR IN WHEN-LINK JOINT CLAUSE.\n");
			                             printf("There is no data in the specified B card field.\n");
			                             printf("field number %d, card type %s\n",
				                             C.conditions[a].fieldB,
				                             error_name);
			                             printf("Line: %d. Exiting.\n",file_lineno);
			                             exit(1);
		                }
				val = strcmp(string1,string2);
				switch(C.conditions[a].comparison) {
					case EQUAL:
						if (!val) continue;
						if (val < 0) {
							skip = 1;
							k=j;
							j = B.list_index;
						}
						if (val > 0) {
							skip = 1;
						}
						break;
					case NOT_EQUAL:
						if (val) continue;
						if (!val) {
							skip = 1;
						}
						break;
					default:
						printf("BAD comparison type in when_else.\n");
						printf("Line %d. Exiting. \n",file_lineno);
						exit(1);
				}
			}
			if (!skip)
				do_links(i,j);
		}
	}
}

void do_links(int a, int b)
{
	int i;
	unsigned char * value;

	printf("\n%% Starting a link\n");
	for (i=0;i<C.link_total;i++) {
		value = link_eval(C.links[i].node,a,b);
		switch(C.links[i].type) 
		{
		case ENTRY:
			if (C.links[i].list == 'A') {
				A.entry_list[a]->fields[C.links[i].field] =
					(unsigned char *) strcpy(my_malloc(20),value);
        			printf("Setting A entry field. Card_type %s , cardnum %s , field number %d , value %s\n",
					A.entry_list[a]->card->name,
					A.entry_list[a]->card->fields[0],
					C.links[i].field,
					value
					);
			}
			if (C.links[i].list == 'B') {
				B.entry_list[b]->fields[C.links[i].field] =
					(unsigned char *) strcpy(my_malloc(20),value);
        			printf("Setting B entry field. Card_type %s , cardnum %s , field number %d , value %s\n",
					B.entry_list[b]->card->name,
					B.entry_list[b]->card->fields[0],
					C.links[i].field,
					value
					);
			}
			break;
		case CARD:
			if (C.links[i].list == 'A') {
				if (A.entries) {
					A.entry_list[a]->card->fields[C.links[i].field] =
						(unsigned char *) strcpy(my_malloc(20),value);
        				printf("Setting A card field. Card_type %s , cardnum %s , field number %d , value %s\n",
						A.entry_list[a]->card->name,
						A.entry_list[a]->card->fields[0],
						C.links[i].field,
						value
						);
						
				} else {
					A.card_list[a]->fields[C.links[i].field] =
						(unsigned char *) strcpy(my_malloc(20),value);
        				printf("Setting A card field. Card_type %s , cardnum %s , field number %d , value %s\n",
						A.card_list[a]->name,
						A.card_list[a]->fields[0],
						C.links[i].field,
						value
						);
				}
			}
			if (C.links[i].list == 'B') {
				if (B.entries) {
					B.entry_list[b]->card->fields[C.links[i].field] =
						(unsigned char *) strcpy(my_malloc(20),value);
        				printf("Setting B card field. Card_type %s , cardnum %s , field number %d , value %s\n",
						B.entry_list[b]->card->name,
						B.entry_list[b]->card->fields[0],
						C.links[i].field,
						value
						);
				} else {
					B.card_list[b]->fields[C.links[i].field] =
						(unsigned char *) strcpy(my_malloc(20),value);
        				printf("Setting B card field. Card_type %s , cardnum %s , field number %d , value %s\n",
						B.card_list[b]->name,
						B.card_list[b]->fields[0],
						C.links[i].field,
						value
						);
				}
			}
			break;
		case STACK:
			if (C.links[i].list == 'A') {
				strcpy(rpl_list[A.rplnum]->fields[C.links[i].field],value);
				printf("Stack field\n");
			}
			if (C.links[i].list == 'B') {
				strcpy(rpl_list[B.rplnum]->fields[C.links[i].field],value);
				printf("Stack field\n");
			}
			break;
		default:
			printf("BAD link type in when_else.\n");
			printf("Line %d. Exiting. \n",file_lineno);
			exit(1);
		}
#if 0
        	printf("\nSetting Link type:%d card %d entry %d set %c\n",
			C.links[i].type,a,b,C.links[i].list);
#endif
	}
	printf("%% Finishing a link\n\n");
}

unsigned char *link_eval( struct LinkNode * node, int a, int b)
{
	unsigned char *left, *right;
	static unsigned char buf[100];
	unsigned char *return_string;

	switch(node->type) 
	{
		case OPERATOR:
			left = link_eval(node->left,a,b);
			right = link_eval(node->right,a,b);
			break;
		case ENTRY:
			if (node->list == 'A') {	
				return_string = A.entry_list[a]->fields[node->field];
				if (!return_string) {
					printf("Something's wrong in link side of when-link.\n");
					printf("I was getting an entry field (double empty brackets)\n");
					printf("but the field has no value\n");
					printf("From the A set.\n");
					printf("Card type %s, number %s , field %d. \n",
						A.entry_list[a]->card->name,
						A.entry_list[a]->card->fields[0],
						node->field);
					printf("Line %d. Exiting.\n",file_lineno);
					exit(0);
				}
				return return_string;
			}
			if (node->list == 'B') {
				return_string = B.entry_list[b]->fields[node->field];
				if (!return_string) {
					printf("Something's wrong in link side of when-link.\n");
					printf("I was getting an entry field (double empty brackets)\n");
					printf("but the field has no value\n");
					printf("From the B set.\n");
					printf("Card type %s, number %s , field %d. \n",
						B.entry_list[b]->card->name,
						B.entry_list[b]->card->fields[0],
						node->field);
					printf("Line %d. Exiting.\n",file_lineno);
					exit(0);
				}
				return return_string;
			}
			printf("Bad case in Link.Line %d. Exiting.\n",file_lineno);
			exit(1);
		case CARD:
			if (node->list == 'A') {
				if (A.entries) {
					return_string = A.entry_list[a]->card->fields[node->field];
					if (!return_string) {
						printf("Something's wrong in link side of when-link.\n");
						printf("I was getting a card field (single empty bracket)\n");
						printf("(one found through an entry-level search)");
						printf("but the field has no value\n");
						printf("From the A set.\n");
						printf("Card type %s, number %s , field %d. \n",
							A.entry_list[a]->card->name,
							A.entry_list[a]->card->fields[0],
							node->field);
						printf("Line %d. Exiting.\n",file_lineno);
						exit(0);
					}
					return return_string;
				} else {
					return_string = A.card_list[a]->fields[node->field];
					if (!return_string) {
						printf("Something's wrong in link side of when-link.\n");
						printf("I was getting a card field (single empty bracket)\n");
						printf("(one found through an card-level search)");
						printf("but the field has no value\n");
						printf("From the A set.\n");
						printf("Card type %s, number %s , field %d. \n",
							A.card_list[a]->name,
							A.card_list[a]->fields[0],
							node->field);
						printf("Line %d. Exiting.\n",file_lineno);
						exit(0);
					}
					return return_string;
				}
			}
			if (node->list == 'B') {
				if (B.entries) {
					return_string = B.entry_list[b]->card->fields[node->field];
					if (!return_string) {
						printf("Something's wrong in link side of when-link.\n");
						printf("I was getting a card field (single empty bracket)\n");
						printf("(one found through an entry-level search)");
						printf("but the field has no value\n");
						printf("From the B set.\n");
						printf("Card type %s, number %s , field %d. \n",
							B.entry_list[b]->card->name,
							B.entry_list[b]->card->fields[0],
							node->field);
						printf("Line %d. Exiting.\n",file_lineno);
						exit(0);
					}
					return return_string;
				} else {
					return_string = B.card_list[b]->fields[node->field];
					if (!return_string) {
						printf("Something's wrong in link side of when-link.\n");
						printf("I was getting a card field (single empty bracket)\n");
						printf("(one found through an card-level search)");
						printf("but the field has no value\n");
						printf("From the B set.\n");
						printf("Card type %s, number %s , field %d. \n",
							B.card_list[b]->name,
							B.card_list[b]->fields[0],
							node->field);
						printf("Line %d. Exiting.\n",file_lineno);
						exit(0);
					}
					return return_string;
				}
			}
			printf("Bad case in Link.Line %d. Exiting.\n",file_lineno);
			exit(1);
		case STACK:
			if (node->list == 'A') {
				return_string = rpl_list[A.rplnum]->fields[node->field];
				return return_string;
			}
			if (node->list == 'B') {
				return_string = rpl_list[B.rplnum]->fields[node->field];
				return return_string;
			}
			printf("Bad case in Link.Line %d. Exiting.\n",file_lineno);
			exit(1);
		case VALUE:
			return node->value;
		default:
			printf("BAD comparison type in when_else.\n");
			printf("Line %d. Exiting. \n",file_lineno);
			exit(1);
	}
	switch(node->operator) {
		case PLUS:	
			return itoa(atoi(left)+atoi(right),buf);
		case MINUS:	
			return itoa(atoi(left)-atoi(right),buf);
		default:
			printf("BAD operator in when_else.\n");
			printf("Line %d. Exiting. \n",file_lineno);
			exit(1);

	}
}

int card_eval(struct Condition *condition, struct Card *card)
{
	int val;

	/* equal or not equal to a field value */
	  if (condition->field==-1) {
	val = strcmp(card->name,
		condition->value);
	  } else {
	val = strcmp(card->fields[condition->field],
		condition->value);
	  }
	switch (condition->comparison) 
	{
		case EQUAL:
			if (val)
				return 0;
			else
				return 1;
		case NOT_EQUAL:
			if (val)
				return 1;
			else
				return 0;
		default:
			printf("BAD comparison type in when_else.\n");
			printf("Line %d. Exiting. \n",file_lineno);
			exit(1);
	}
}

int entry_eval(struct Condition *condition, struct Entry *entry)
{
	/* equal or not equal to a field value */
	int val;

	/* equal or not equal to a field value */
	val = strcmp(entry->fields[condition->field],
		condition->value);
	switch (condition->comparison) 
	{
		case EQUAL:
			if (val)
				return 0;
			else
				return 1;
		case NOT_EQUAL:
			if (val)
				return 1;
			else
				return 0;
		default:
			printf("BAD comparison type in when_else.\n");
			printf("Line %d. Exiting. \n",file_lineno);
			exit(1);
	}
}

/* afterwards, we clean up by resetting the trees and other structures
	for passsing up information to this level. */
void clean_a_b()
{
	int i;

	/* free lists */
	if (A.entries)
		free(A.entry_list);
	else
		free(A.card_list);

	if (B.entries)
		free(B.entry_list);
	else
		free(B.card_list);

	/* Walk C links */
	for (i=0;i<20;i++) {
		node_free(C.links[i].node);
	}

	/* zero A and B */
	memset(&A,0,sizeof(A));
	memset(&B,0,sizeof(B));
	memset(&C,0,sizeof(C));

	A.rplnum = -1;
	B.rplnum = -1;
}

node_free(struct LinkNode *node)
{
	if (!node) return;
	node_free(node->left);
	node_free(node->right);
	my_free(node);
}

int A_card_lookup(unsigned char *field)
{
	int i;

	for (i=0;i<A.card_condition_total;i++){
		if (A.card_condition[i].field == -1) {
			return card_field_index(field,A.card_condition[i].value);		
		}
	}
	printf("Couldn't find card_type in A conditions for field %s. Exiting. \n",field);
	printf("Check WHEN/LINK: Make sure (A)Card[..][] == \"CARD_TYPE\" comparision is done first.\n");
	exit(1);

}
	
int A_entry_lookup(unsigned char *field)
{
	int i;

	for (i=0;i<A.card_condition_total;i++){
		if (A.card_condition[i].field == -1) {
			return entry_field_index(field,A.card_condition[i].value);		
		}
	}
	printf("Couldn't find entry_type in A conditions for field %s. Exiting. \n",field);
	printf("Check WHEN/LINK: Make sure (A)Card[..][] == \"CARD_TYPE\" comparision is done first.\n");
	exit(1);
}

int B_card_lookup(unsigned char *field)
{
	int i;

	for (i=0;i<B.card_condition_total;i++){
		if (B.card_condition[i].field == -1) {
			return card_field_index(field,B.card_condition[i].value);		
		}
	}
	printf("Couldn't find card_type in B conditions for field %s. Exiting. \n",field);
	printf("Check WHEN/LINK: Make sure (B)Card[..][] == \"CARD_TYPE\" comparision is done first.\n");
	exit(1);
}

int B_entry_lookup(unsigned char *field)
{
	int i;

	for (i=0;i<B.card_condition_total;i++){
		if (B.card_condition[i].field == -1) {
			return entry_field_index(field,B.card_condition[i].value);		
		}
	}
	printf("Couldn't find entry_type in B conditions for field %s. Exiting. \n",field);
	printf("Check WHEN/LINK: Make sure (B)Card[..][] == \"CARD_TYPE\" comparision is done first.\n");
	exit(1);
}

make_new_card()
{
	int i;
	unsigned char buf[1000];

	/* create a new card */
	current_card = (struct Card *) my_malloc(sizeof(struct Card));
	current_card->fields[0] = (unsigned char *) 0;

	/* create its initial entry */
	entry_index = 0;
	current_card->entries = (struct Entry **) my_malloc(100*sizeof(struct Entry *));
	current_card->entries[entry_index] = (struct Entry *) my_malloc(sizeof(struct Entry));
	current_entry = current_card->entries[entry_index];
	current_entry->card = current_card;
	entry_index++;

	/* add card to list of cards for this stack */
	rpl_list[current_rpl]->cards[rpl_list[current_rpl]->card_index] = current_card; 
	rpl_list[current_rpl]->card_index++;
	/* make that list longer if necessary */
	if (!(rpl_list[current_rpl]->card_index%CARD_PTR_BLOCK)) {
		int blocks, size;
		blocks = rpl_list[current_rpl]->card_index/CARD_PTR_BLOCK;
		size = (blocks+1)*CARD_PTR_BLOCK*(sizeof(struct Card *));
		if ((rpl_list[current_rpl]->cards = (struct Card **) realloc(rpl_list[current_rpl]->cards,size))==NULL) {
			printf("REALLOC FAILS. Exiting.\n");
			exit(1);
		}
		memset(&rpl_list[current_rpl]->cards[rpl_list[current_rpl]->card_index],
			0,CARD_PTR_BLOCK*(sizeof(struct Card *)));
	}
#	ifdef SD_DEBUG3
	printf("NEW_CARD %d\n",rpl_list[current_rpl]->card_index);
#	endif

	/* increment CARDNUM */
	u_card_num++;
	set_local_value("CARDNUM",itoa(u_card_num,buf));

	/* make sure this agrees with OUR cardnum */
	if (u_card_num != rpl_list[current_rpl]->card_index-1) {
		printf("Card numbers out of sync. Internal error. \n");
		printf("Line %d. Exiting.\n",file_lineno);
		exit(1);
	}

	/* if in given, add to skip_list for undo. The extensible array doesn't matter. */
	if (current_given) {
		for (i=0; i<current_given; i++) {
			skip_list[i][skip_list_index[i]++] = (unsigned char *) current_card;
			skip_list[i][skip_list_index[i]++] = (unsigned char *) current_card->entries;
			skip_list[i][skip_list_index[i]++] = (unsigned char *) current_entry;
		}
	}
}

make_new_entry()
{
	int i;

	if (current_entry) current_entry->used = 1;
	current_entry = (struct Entry *) my_malloc(sizeof(struct Entry));
	current_card->entries[entry_index] = current_entry;
	current_entry->card = current_card;
	entry_index++;

	/* if in given, add to skip_list for undo */
	if (current_given) {
		for (i=0; i<current_given; i++) {
			skip_list[i][skip_list_index[i]++] = (unsigned char *) current_entry;
		}
	}
}

lose_entry()
{
	int i,j;

	entry_index--;
	current_card->entries[entry_index] = 0;
	
	if (current_given) {
		for (i=0; i<current_given; i++) {
			for (j=0;j<skip_list_index[i];j++)  {
				if (skip_list[i][j] == (unsigned char *) current_entry) 
					skip_list[i][j] = 0;
			}
		}
	}
	my_free(current_entry);
	if(!entry_index)
	  current_entry = (struct Entry *) NULL;
        else
	  current_entry = current_card->entries[entry_index-1];
}

given_undo()
{
	int i;
	unsigned char buf[1000];

	if (current_given) {
		/* the card/entries/entry triplet we'll reset to is not stored on this list */
		/* undo this given */
		for (i=0; i<skip_list_index[current_given-1]; i++) {
			if (skip_list[current_given-1][i]) my_free(skip_list[current_given-1][i]);
		}
		/* delete that many values from the bottom of earlier skip lists */
		for (i=0; i<(current_given-1); i++ ) {
			skip_list_index[i]-=skip_list_index[current_given-1];	
		}
		skip_list_index[current_given-1] = 0;
	
		/* reset that card to it's original state (which SHOULD be blank) */
		current_card = skip_card[current_given-1];
		memset(current_card,0,sizeof(struct Card));
	
		current_card->fields[0] = (unsigned char *) 0;
		entry_index = 0;
	
		current_card->entries = skip_entries[current_given-1];
		memset(current_card->entries,0,(100*sizeof(struct Entry *)));
	
		current_card->entries[entry_index] = skip_entry[current_given-1];
		memset(current_card->entries[entry_index],0,(sizeof(struct Entry)));
	
		current_entry = current_card->entries[entry_index];
		current_entry->card = current_card;
		entry_index++;
		rpl_list[current_rpl]->card_index = skip_card_index[current_given-1] - 1;
		rpl_list[current_rpl]->cards[rpl_list[current_rpl]->card_index] = current_card; 
		rpl_list[current_rpl]->card_index++;
		u_card_num = skip_card_num[current_given-1];
		set_local_value("CARDNUM",itoa(u_card_num,buf));

		ptl_index = 0;
		memset(page_total_list,0,sizeof(page_total_list));
	}
}
clear_skip_list()
{
	/* Reset skip index for current given */
	skip_list_index[current_given-1] = 0;
}

set_u_page_total(stack_block *block)
{
	int i, j, k, len, l_index;
	unsigned char buf[1000], field_name_list[10][50], field_value_list[10][100], *ptr;
	int entry_limit, entry_count, field_index, starting_point;

	memset(field_name_list,0,sizeof(field_name_list));
	memset(field_value_list,0,sizeof(field_value_list));

	entry_limit = get_entry_limit(block);
	max_book_entries = entry_limit;

	/* an entry_limit of -1 indicates that the maximum number of entries 
		was not defined by the user. */
	if (max_book_entries == -1) 
	{
		set_local_value("PAGETOTAL",itoa(max_book_entries,buf));
		return;
	}

	/* items on the line we get from given_list and book_list
		are separated by blanks */
	len = strlen(block->value);
	l_index = 0;
	j=0;
	for (i=0; block->value[i] ; i++) {
		if (block->value[i] == '$') continue;
		if (block->value[i] == ' ') {
			field_name_list[l_index][j] = '\0';
			j=0;
			l_index++;
			continue;
		} else {
			field_name_list[l_index][j] = block->value[i];
			j++;
		}
	}
	field_name_list[l_index][0] = '\0'; /* clear number */
	
	/* if in given, start at given data index. Else 0. */
	if (current_given) {
		entry_count = 1;
		starting_point = given_data_index[current_given-1];
	/* get initial values */
		for (j=0; strlen(field_name_list[j]); j++) {
			field_index = data_field_index(field_name_list[j]);
			strcpy(field_value_list[j],(ptr=data_field( starting_point, field_index)));
			my_free(ptr);
		}
		for (i=starting_point; within_given_range(i,current_given-1); i++) {
			for (j=0; strlen(field_name_list[j]); j++) {
				field_index = data_field_index(field_name_list[j]);
				strcpy(buf,(ptr=data_field( i, field_index)));
				my_free(ptr);
				if (strcmp(field_value_list[j],buf)) {
					entry_count++;
					for (k=0; strlen(field_name_list[k]); k++) {
						field_index = data_field_index(field_name_list[k]);
						strcpy(field_value_list[k],(ptr=data_field( i, field_index)));
						my_free(ptr);
					}
					break;
				}
			}
		}
		u_page_total = (entry_count + (entry_limit - 1))/ entry_limit;
	} else {
		entry_count = 1;
		starting_point = 0; 
		/* get initial values */
		for (j=0; strlen(field_name_list[j]); j++) {
			field_index = data_field_index(field_name_list[j]);
			strcpy(field_value_list[j],(ptr=data_field( starting_point, field_index)));
			my_free(ptr);
		}
		for (i=starting_point; real_data(i); i++) {
			for (j=0; strlen(field_name_list[j]); j++) {
				field_index = data_field_index(field_name_list[j]);
				strcpy(buf,(ptr=data_field( i, field_index)));
				my_free(ptr);
				if (strcmp(field_value_list[j],buf)) {
					entry_count++;
					for (k=0; strlen(field_name_list[k]); k++) {
						field_index = data_field_index(field_name_list[k]);
						strcpy(field_value_list[k],(ptr=data_field( i, field_index)));
						my_free(ptr);
					}
					break;
				}
			}
		}
		u_page_total = (entry_count + (entry_limit - 1))/ entry_limit;
	}
	set_local_value("PAGETOTAL",itoa(u_page_total,buf));
}

/* you have to give within_given_range every index from 
	given_data_index[current_range-1] so that you can see it fail. */
int within_given_range(int index, int given)
{
	int j,k,field_index;
	unsigned char *ptr, buf[1000];

#				ifdef SD_DEBUG3
				printf("S:WGR\n");
#				endif
	if (index > data_records) 
	{
#				ifdef SD_DEBUG3
				printf("F:WGR\n");
#				endif
		return 0;
	}

	for (k=given; k >= 0; k--) {
		for (j=0; strlen(given_field_name_list[k][j]); j++) {
			field_index = data_field_index(given_field_name_list[k][j]);
			strcpy(buf,(ptr=data_field( index, field_index)));
			my_free(ptr);
			if (strcmp(given_field_value_list[k][j],buf)) {
#				ifdef SD_DEBUG3
				printf("F:WGR\n");
#				endif
				return 0;
			}
		}
	}

#				ifdef SD_DEBUG3
				printf("F:WGR\n");
#				endif
	return 1;
}

int real_data(int index)
{
	unsigned char *ptr1, *ptr2;

	ptr1 = (unsigned char *) 0;
	ptr2 = (unsigned char *) 0;
	if (!strlen(ptr1=data_field(index,0)) && !strlen(ptr2=data_field(index,1))) 
	{
		my_free(ptr1);
		my_free(ptr2);
		return 0;
	}
	if (ptr1) my_free(ptr1);
	if (ptr2) my_free(ptr2);
	return 1;
}

int get_entry_limit(stack_block *block)
{
	int len, i, j;
	unsigned char buf[1000];

	len = strlen(block->value);
	for (i=len; block->value[i] != ' ' && i > 0; i--);
	for (j=0;i<len;i++,j++) buf[j] = block->value[i];
	buf[j] = '\0';
	return atoi(buf);
}

int given_advance(stack_block *block)
{
	int i, j, k, field_index;
	unsigned char *ptr, buf[1000];

	starting_point = given_data_index[current_given-1];
	for (i=starting_point;1; i++) {
		if (current_given-1 > 0) {
			if (!within_given_range(i,current_given-2)) return 0;	
		}
		if (!real_data(i)) return 0;
		
		for (j=0; strlen(given_field_name_list[current_given-1][j]); j++) {
			field_index = data_field_index(given_field_name_list[current_given-1][j]);
			strcpy(buf,(ptr=data_field( i, field_index)));
			my_free(ptr);
			if (strcmp(given_field_value_list[current_given-1][j],buf)) {
				for (k=0; strlen(given_field_name_list[current_given-1][k]); k++) {
					field_index = data_field_index(given_field_name_list[current_given-1][k]);
					strcpy(given_field_value_list[current_given-1][k],(ptr=data_field( i, field_index)));
					my_free(ptr);
				}
				given_data_index[current_given-1] = i;
				return 1;
			}
		}
	}
}

void init_given_fields(stack_block *block)
{
	int len, i, j, k, field_index, l_index;
	unsigned char *ptr;

	len = strlen(block->value);
	l_index = 0;
	j=0;
	for (i=0; block->value[i] ; i++) {
		if (block->value[i] == '$') continue;
		if (block->value[i] == ' ') {
			given_field_name_list[current_given][l_index][j] = '\0';
			j=0;
			l_index++;
			continue;
		} else {
			given_field_name_list[current_given][l_index][j] = block->value[i];
			j++;
		}
	}
	given_field_name_list[current_given][l_index][j] = '\0';
	given_field_name_list[current_given][l_index+1][0] = '\0';
	for (k=0; strlen(given_field_name_list[current_given][k]); k++) {
		field_index = data_field_index(given_field_name_list[current_given][k]);
		strcpy(given_field_value_list[current_given][k],(ptr=data_field( given_data_index[current_given], field_index)));
		my_free(ptr);
	}
}

void init_cb_fields(stack_block *block)
{
	int len, i, j, k, field_index, l_index;
	unsigned char *ptr;

	len = strlen(block->value);
	l_index = 0;
	j=0;
	for (i=0; block->value[i] ; i++) {
		if (block->value[i] == '$') continue;
		if (block->value[i] == ' ') {
			cb_field_name_list[l_index][j] = '\0';
			j=0;
			l_index++;
			continue;
		} else {
			cb_field_name_list[l_index][j] = block->value[i];
			j++;
		}
	}
	cb_field_name_list[l_index][0] = '\0'; /* clear number */
	for (k=0; strlen(cb_field_name_list[k]); k++) {
		field_index = data_field_index(cb_field_name_list[k]);
		strcpy(cb_field_value_list[k],(ptr=data_field( card_book_data_index, field_index)));
		my_free(ptr);
	}
}


int card_book_advance()
{
	unsigned char buf[100000], *ptr;
	int field_index =0, starting_point =0;
	int i =0,j =0,k =0;

	/* if in given, start at given data index. Else 0. */
	if (current_given) {
		starting_point = card_book_data_index;
		for (i=starting_point; within_given_range(i, current_given-1); i++) {
			for (j=0; strlen(cb_field_name_list[j]); j++) {
				field_index = data_field_index(cb_field_name_list[j]);
				strcpy(buf,(ptr=data_field( i, field_index)));
				my_free(ptr);
				if (strcmp(cb_field_value_list[j],buf)) {
					for (k=0; strlen(cb_field_name_list[k]); k++) {
						field_index = data_field_index(cb_field_name_list[k]);
						strcpy(cb_field_value_list[k],(ptr=data_field( i, field_index)));
						my_free(ptr);
					}
					card_book_data_index = i;
					/* Set Global User variable DATA_INDEX */
					set_local_value("DATA_INDEX",itoa(card_book_data_index,buf));
					return 1;
				}
			}
		}
	} else {
		starting_point = card_book_data_index;
		for (i=starting_point; real_data(i); i++) {
			for (j=0; strlen(cb_field_name_list[j]); j++) {
				field_index = data_field_index(cb_field_name_list[j]);
				strcpy(buf,(ptr=data_field( i, field_index)));
				my_free(ptr);
				if (strcmp(cb_field_value_list[j],buf)) {
					for (k=0; strlen(cb_field_name_list[k]); k++) {
						field_index = data_field_index(cb_field_name_list[k]);
						strcpy(cb_field_value_list[k],(ptr=data_field( i, field_index)));
						my_free(ptr);
					}
					card_book_data_index = i;
					/* Set Global User variable DATA_INDEX */
					set_local_value("DATA_INDEX",itoa(card_book_data_index,buf));
					return 1;
				}
			}
		}
	}
	return 0;

}

make_and_push_card()
{
	pushed_current_card = current_card;
	pushed_current_entry = current_entry;
	pushed_u_entry_total = u_entry_total;
	pushed_entry_index = entry_index;

	/* do not push any rpl_list information.
		This is maintained independent of
		current_card & co. above. 
	*/

	make_new_card();
}

finish_and_pop_card()
{
	unsigned char buf[100];

	current_card = pushed_current_card;
	current_entry = pushed_current_entry;
	u_entry_total = pushed_u_entry_total;
	set_local_value("ENTRYTOTAL",itoa(u_entry_total,buf));
	entry_index = pushed_entry_index;
}


